"""Solution objects for beam propagations."""

import logging
import typing

import numpy as np

from finesse.components import Cavity, Space
from finesse.components.general import Connector
from finesse.components.node import Port, OpticalNode
from finesse.environment import is_interactive
from finesse.gaussian import BeamParam
from finesse.symbols import evaluate, Symbol
from finesse.solutions.base import BaseSolution
from finesse.utilities.misc import is_iterable, pairwise

LOGGER = logging.getLogger(__name__)


class ABCDSolution(BaseSolution):
    """Solution for a composite ABCD calculation.

    Parameters
    ----------
    M : array-like or two-element tuple of array-like
        The ABCD matrix / matrices.

    direction : str
        Direction / plane of computation.

        ``"both"`` indicates that `M` is a tuple of the ABCD matrices
        computed over both the tangential and sagittal planes.

        ``"x"`` implies M is the ABCD matrix computed over the tangential plane.

        ``"y"`` implies M is the ABCD matrix computed over the sagittal plane.

    symbolic : bool
        Flag indicating whether the calculations are symbolic.
    """

    def __init__(self, name, M, direction, symbolic):
        super().__init__(name)
        self.empty = False  # tree drawing fill circle

        self._M = M
        self.__direction = direction
        self.__symbolic = symbolic

    def __str__(self):
        return str(self._M)

    @property
    def direction(self):
        """The plane in which this ABCD matrix was computed - 'x'
        for tangential, 'y' for sagittal.

        :getter: Returns the ABCD matrix plane.
        """
        return self.__direction

    @property
    def symbolic(self):
        """Indicates whether this ABCD solution is symbolic.

        :getter: Returns `True` if this stores symbolic expressions, `False` otherwise.
        """
        return self.__symbolic

    @property
    def M(self):
        """A copy of the underlying ABCD matrix as a :class:`numpy.ndarray`.

        :getter: Returns a copy of underlying ABCD matrix.
        """
        return self._M.copy()

    def eval(self):
        """Evaluate the symbolic ABCD matrix.

        Computes the numeric form of the ABCD matrix using
        the ``eval`` method of each parameter reference.

        Returns
        -------
        out : :class:`numpy.ndarray`
            A numeric matrix for the evaluated ABCD.
        """
        if not self.symbolic:
            return self.M

        return evaluate(self.M)


class PropagationSolution(BaseSolution):
    """Solution representation of a call to
    :func:`~finesse.tracing.tools.propagate_beam`.

    This class contains useful attributes and methods for accessing properties of the
    beam that was propagated through the path specified by the above function call. If
    this propagation call was symbolic then each property returned from this class will
    also be symbolic - evaluate these using the ``eval`` method of the symbolic
    expression.

    Note that PropagationSolution objects are returned via
    :func:`~finesse.tracing.tools.propagate_beam` (or :meth:`.Model.propagate_beam`),
    they should never need to be created manually.

    See :ref:`propagating_beams` for details and examples on using this class.
    """

    def __init__(self, name, node_info, comp_info, symbolic):
        super().__init__(name)
        self.node_info = node_info
        self.comp_info = comp_info
        self.__symbolic = symbolic

        self.total_acc_gouy = 0
        for info in comp_info.values():
            self.total_acc_gouy += info.get("acc_gouy", 0)

        self.empty = False

        # NOTE (sjr) For non-symbolic propagations, we need to store the fixed
        #            space length as it was at the time of the propagate_beam
        #            call to avoid any issues (when plotting the solution) if a
        #            space length is changed in the model afterwards
        if not self.symbolic:
            self.__frozen_space_lengths = {
                space: evaluate(space.L.value) for space in self.spaces
            }
        else:
            self.__frozen_space_lengths = {}

    def __get_model(self):
        nodes = list(self.node_info.keys())
        if not nodes:
            return None

        return nodes[0]._model

    def __get_component(self, x):
        if isinstance(x, str):
            name = x
            model = self.__get_model()
            if model is None:
                raise RuntimeError("Empty propagationSolution - no path traversed!")

            x = model.elements.get(name)
            if x is None:
                raise ValueError(f"No component of name {name} exists in the model")

        if not isinstance(x, Connector):
            raise TypeError(
                "Invalid type for component. Expected Connector " f"but got {type(x)}"
            )

        if x not in self.comp_info:
            raise ValueError(
                f"No component of name {x.name} exists in the propagated path"
            )

        return x

    @property
    def symbolic(self):
        """Whether the propagation solution is symbolic."""
        return self.__symbolic

    @property
    def start_node(self):
        """The starting node of the propagation."""
        nodes = list(self.node_info.keys())
        if not nodes:
            return None

        return nodes[0]

    @property
    def end_node(self):
        """The final node of the propagation."""
        nodes = list(self.node_info.keys())
        if not nodes:
            return None

        return nodes[-1]

    @property
    def nodes(self):
        """A list of all the nodes traversed, in order."""
        return list(self.node_info.keys())

    @property
    def ports(self):
        """A list of all the ports traversed, in order."""
        return list(dict.fromkeys(node.port for node in self.nodes))

    @property
    def spaces(self):
        """A list of all spaces traversed, in order."""
        space_entries = dict(
            filter(lambda x: "acc_gouy" in x[1], self.comp_info.items())
        )
        return list(space_entries.keys())

    @property
    def components(self):
        """A list of all components (excluding spaces) traversed, in order."""
        return list(filter(lambda x: not isinstance(x, Space), self.comp_info.keys()))

    @property
    def positions(self):
        """A dictionary of the :class:`.Connector` instances to their positions
        (relative to the start node)."""
        pos = {}
        for node, info in self.node_info.items():
            if node.component in pos:
                continue

            pos[node.component] = info["z"]

        return pos

    @property
    def path_length(self):
        """The geometric path length of the traversed path.

        Equal to the sum of each space length in the path.
        """
        nodes = list(self.node_info.keys())
        if not nodes:
            return 0

        return self.node_info[nodes[-1]]["z"]

    @property
    def optical_path_length(self):
        """The optical path length of the traversed path.

        Equal to the sum of the product of each space length and refractive index in the
        path.
        """
        nodes = self.nodes
        if not nodes:
            return 0

        return self.node_info[nodes[-1]]["z_optical"]

    @property
    def beamsizes(self):
        """Dictionary of node to beam size mappings."""
        return {node: info["q"].w for node, info in self.node_info.items()}

    @property
    def ws(self):
        """Identical to :attr:`PropagationSolution.beamsizes`"""
        return self.beamsizes

    @property
    def waistpositions(self):
        """Dictionary of node to waist position (as measured from node) mappings."""
        return {node: info["q"].z for node, info in self.node_info.items()}

    @property
    def z0s(self):
        """Identical to :attr:`PropagationSolution.waistpositions`"""
        return self.waistpositions

    @property
    def qs(self):
        """Dictionary of node to beam parameter mappings."""
        return {node: info["q"] for node, info in self.node_info.items()}

    @property
    def full_ABCD(self):
        """The full, composite ABCD matrix from the start to the end of the path."""
        last_node_info = list(self.node_info.values())[-1]
        return last_node_info["ABCD"]

    def acc_gouy(self, *args):
        """Accumulated Gouy phase over a sequence of spaces.

        Parameters
        ----------
        args : sequence of args
            Space components or names of spaces.

        Returns
        -------
        agouy : float or :class:`.Operation`
            The accumulated Gouy phase over the given spaces.
        """
        agouy = 0
        model = self.__get_model()
        if model is None:
            return agouy

        for space in args:
            if isinstance(space, str):
                space = model.elements.get(space)

            agouy += self.comp_info[space]["acc_gouy"]
        return agouy

    def acc_gouy_up_to(self, point):
        """Accumulated Gouy phase up to a `point` in the traversed path.

        This computes the cumulative Gouy phase from the
        :attr:`.PropagationSolution.start_node` to the specified `point`.

        Parameters
        ----------
        point : :class:`.OpticalNode` or :class:`.Port` or :class:`.Connector` or str
            A node, port, component or name of component up to which to compute
            the accumulated Gouy phase.

        Returns
        -------
        agouy : float or :class:`.Operation`
            The accumulated Gouy phase up to the given `point`.
        """

        if isinstance(point, (OpticalNode, Port)):
            point = point.component

        point = self.__get_component(point)

        agouy = 0
        for comp, info in self.comp_info.items():
            if comp == point:
                break

            agouy += info.get("acc_gouy", 0)
        return agouy

    def abcd(self, up_to):
        """Composite ABCD matrix up to a specific point in the path.

        Parameters
        ----------
        up_to : :class:`.OpticalNode` or :class:`.Port`
            The location in the path at which to get the composite ABCD matrix. This
            can be an optical node or a port.

        Returns
        -------
        M : :class:`numpy.ndarray`
            The ABCD matrix, as a NumPy array, computed up to the specified location.

        Raises
        ------
        ke : KeyError
            If `up_to` is a node or port which does not exist within the solution.

        te : TypeError
            If `up_to` is not an :class:`.OpticalNode` or :class:`.Port`.
        """
        if isinstance(up_to, OpticalNode):
            if up_to not in self.node_info:
                raise KeyError(
                    f"No optical node of name {up_to.full_name} in the solution."
                )

            M = self.node_info[up_to]["ABCD"]

        elif isinstance(up_to, Port):
            ientry = self.node_info.get(up_to.i, {})
            oentry = self.node_info.get(up_to.o, {})

            if not ientry and not oentry:
                raise KeyError(f"No port of name {up_to.full_name} in the solution.")

            in_M = ientry.get("ABCD", None)
            out_M = oentry.get("ABCD", None)

            if in_M is not None:
                M = in_M
            elif out_M is not None:
                M = out_M
            else:
                raise RuntimeError(
                    "Bug encountered! No composite ABCD matrix entry found "
                    f"at the port {up_to.full_name}."
                )

        ## TODO (sjr) Add support for up_to being a Connector (or name of Connector)

        else:
            raise TypeError(f"Invalid type for up_to: {up_to}")

        return M

    def q(self, at, which=None):
        """Beam parameter at a given location of the path.

        Parameters
        ----------
        at : :class:`.OpticalNode` or :class:`.Port` or :class:`.Connector` or str
            The location in the path at which to get the beam parameter. This can
            be an optical node, a port, a connector or the name of a connector.

        which : str, optional; default: None
            The direction of the beam parameter to retrieve. Only used if `at`
            is a connector or name of a connector. Valid options are "in" for
            the input node of the connector, or "out" for the output node of
            the connector.

        Returns
        -------
        q : :class:`.BeamParam`
            If `at` is an :class:`.OpticalNode` instance, or a :class:`.Connector`
            object *and* `which` was specified, then the beam parameter corresponding
            to this node is returned.

        qs : dict
            Otherwise, a dictionary of the input and output node mappings to their
            respective beam parameters is returned.

        Raises
        ------
        ke : KeyError
            If `at` is a node or port which does not exist within the solution.

        ve : ValueError
            If `which` is an invalid argument (see parameters above). Or if `at` corresponds
            to a non-existent connector (or a connector not in the traced path).

        te : TypeError
            If `at` is not an :class:`.OpticalNode`, :class:`.Port`, :class:`.Connector`
            or string.
        """
        if isinstance(at, OpticalNode):
            if at not in self.node_info:
                if at.opposite in self.node_info:
                    # If the opposite node direction is in the solution
                    # then return the reverse of this beam parameter at
                    # this node (i.e. -q*)
                    q = self.node_info[at.opposite]["q"].reverse()
                else:
                    raise KeyError(
                        f"No optical node of name {at.full_name} (nor its opposite) "
                        f"in the solution."
                    )
            else:
                q = self.node_info[at]["q"]

        elif isinstance(at, Port):
            ientry = self.node_info.get(at.i, {})
            oentry = self.node_info.get(at.o, {})

            if not ientry and not oentry:
                raise KeyError(f"No port of name {at.full_name} in the solution.")

            in_q = ientry.get("q", None)
            out_q = oentry.get("q", None)

            if in_q is None:
                in_q = out_q.reverse()
            elif out_q is None:
                out_q = in_q.reverse()

            q = {at.i: in_q, at.o: out_q}

        elif isinstance(at, Connector) or isinstance(at, str):
            at = self.__get_component(at)

            if which is None:
                q = self.comp_info[at]

            elif isinstance(which, str):
                if which.casefold() == "in":
                    nodes = self.comp_info[at].keys()
                    in_node = list(filter(lambda n: n.is_input, nodes))
                    if not in_node:
                        raise ValueError(
                            f"No input node entry of component {at.name} exists"
                        )
                    q = self.comp_info[at][in_node[0]]
                elif which.casefold() == "out":
                    nodes = self.comp_info[at].keys()
                    in_node = list(filter(lambda n: not n.is_input, nodes))
                    if not in_node:
                        raise ValueError(
                            f"No output node entry of component {at.name} exists"
                        )
                    q = self.comp_info[at][in_node[0]]
                else:
                    raise ValueError(
                        "Invalid string value for argument which. Expected 'in' or "
                        f"'out' but got {which}"
                    )

            else:
                raise ValueError(
                    f"Invalid value and type for argument which. Expected 'in' or "
                    f"'out' but got {which}"
                )
        else:
            raise TypeError(f"Invalid type for at: {at}")

        return q

    def beamsize(self, at, which=None):
        """Beam radius at a given location of the path.

        Parameters
        ----------
        See :meth:`.PropagationSolution.q`.

        Returns
        -------
        w : float or :class:`.Operation`
            If `at` is an :class:`.OpticalNode` instance, or a :class:`.Connector`
            object *and* `which` was specified, then the beam size corresponding
            to this node is returned.

        ws : dict
            Otherwise, a dictionary of the input and output node mappings to their
            respective beam sizes is returned.

        Raises
        ------
        See :meth:`.PropagationSolution.q`.
        """
        q_ = self.q(at, which=which)
        # Both nodes present, so return dict of beamsizes for each
        if isinstance(q_, dict):
            return {node: q.w for node, q in q_.items()}

        return q_.w

    def w(self, at, which=None):
        """Identical to :meth:`PropagationSolution.beamsize`."""
        return self.beamsize(at, which=which)

    def waistsize(self, at, which=None):
        """Waist radius as measured from a given location of the path.

        Parameters
        ----------
        See :meth:`.PropagationSolution.q`.

        Returns
        -------
        w : float or :class:`.Operation`
            If `at` is an :class:`.OpticalNode` instance, or a :class:`.Connector`
            object *and* `which` was specified, then the waist size using the beam
            parameter basis at this node is returned.

        ws : dict
            Otherwise, a dictionary of the input and output node mappings to their
            respective waist sizes (using the node beam parameter bases) is returned.

        Raises
        ------
        See :meth:`.PropagationSolution.q`.
        """
        q_ = self.q(at, which=which)
        # Both nodes present, so return dict of beamsizes for each
        if isinstance(q_, dict):
            return {node: q.w0 for node, q in q_.items()}

        return q_.w0

    def w0(self, at, which=None):
        """Identical to :meth:`PropagationSolution.waistsize`."""
        return self.waistsize(at, which=which)

    def waistpos(self, from_point, which=None):
        """Waist position as measured at `from_point`.

        Parameters
        ----------
        See :meth:`.PropagationSolution.q`.

        Returns
        -------
        w : float or :class:`.Operation`
            If `at` is an :class:`.OpticalNode` instance, or a :class:`.Connector`
            object *and* `which` was specified, then the distance to the waist
            from this node is returned.

        ws : dict
            Otherwise, a dictionary of the input and output node mappings to their
            respective distances to the waist is returned.

        Raises
        ------
        See :meth:`.PropagationSolution.q`.
        """
        q_ = self.q(from_point, which=which)
        if isinstance(q_, dict):
            return {node: q.z for node, q in q_.items()}

        return q_.z

    def z0(self, from_point, which=None):
        """Identical to :meth:`PropagationSolution.waistpos`."""
        return self.waistpos(from_point, which=which)

    def __getitem__(self, key):
        if isinstance(key, OpticalNode):
            return self.node_info[key]

        if isinstance(key, Connector) or isinstance(key, str):
            key = self.__get_component(key)
            return self.comp_info[key]

        raise TypeError(f"Invalid type for key: {key}")

    def compute_distances_matrix(self, ztype="geometric", subs=None):
        """Compute the distances between each optic, relative to each other.

        Returns a dict of dicts for each "delta z". Note that each distance value is
        in metres. Use :meth:`PropagationSolution.print_distances_matrix` to print a tabulated
        string representation of this dict.

        Parameters
        ----------
        ztype : str, optional; default: "geometric"
            Type of distance, can be either 'geometric' or 'optical'. In the former case
            the values are the distances between each optic in terms of sums of space lengths
            of each space between them. In the latter case, each value is instead the optical
            path length between each component, i.e. the sum of the product of the space
            length and refractive index of each space between them.

        subs : dict, optional
            A dictionary of model parameter to value substitutions
            to pass to the ``eval`` methods of symbolic expressions.

            If this solution object is not symbolic then this argument
            is ignored.

        Returns
        -------
        deltas : dict
            Dict of dicts for each dz between components.
        """
        if not self.symbolic and subs is not None:
            LOGGER.warning(
                "Ignoring subs=%s kwarg as PropagationSolution is non-symbolic.", subs,
            )

        deltas = {}
        outer_done = set()
        for node1, info1 in self.node_info.items():
            comp1 = node1.component
            if comp1 in outer_done:
                continue

            inner_done = set()
            deltas[comp1] = {}
            for node2, info2 in self.node_info.items():
                comp2 = node2.component
                if comp2 in inner_done:
                    continue

                z2 = self._eval_z_for_display(info2, ztype=ztype, subs=subs)
                z1 = self._eval_z_for_display(info1, ztype=ztype, subs=subs)
                deltas[comp1][comp2] = z2 - z1
                inner_done.add(comp2)

            outer_done.add(comp1)

        return deltas

    def print_distances_matrix(
        self, scale=1, ztype="geometric", tablefmt="fancy_grid", subs=None
    ):
        """Print the distances between each optic, relative to each other, in a table.

        Note that all distances are in metres by default. You can apply a scaling factor,
        via the `scale` argument, to all values to change the units accordingly.

        Parameters
        ----------
        ztype : str, optional; default: "geometric:
            See :meth:`.PropagationSolution.compute_distances_matrix`.

        scale : float, optional
            Scaling factor to apply to distances. Defaults to unity such
            that all distances are in metres.

        tablefmt : str, optional; default: "fancy_grid"
            The plain-text table format to pass to tabulate.

        subs : dict, optional
            A dictionary of model parameter to value substitutions
            to pass to the ``eval`` methods of symbolic expressions.

            If this solution object is not symbolic then this argument
            is ignored.
        """
        from tabulate import tabulate

        deltas = self.compute_distances_matrix(ztype=ztype, subs=subs)

        table = []
        headers = [comp.name for comp in self.components]

        for comp1 in deltas:
            entry = [comp1.name]
            for dz in deltas[comp1].values():
                entry.append(scale * dz)
            table.append(entry)

        print(tabulate(table, headers, tablefmt=tablefmt))

    def position(self, point, ptype="geometric"):
        """Gets the position of the specified `point` relative to the start node.

        Parameters
        ----------
        point : :class:`.OpticalNode` or :class:`.Port` or :class:`.Connector` or str
            The location in the path from which to obtain the relative position. This can
            be an optical node, a port, a connector or the name of a connector.

        ptype : str
            Type of distance, can be either 'geometric' or 'optical'. In the former case
            the value returned is the distance from the start node to `point` as a sum
            of each space length. In the latter case the value returned will be the optical
            path length from the start node to `point`, i.e. a sum of each space length
            multiplied by its refractive index.

        Returns
        -------
        z : float or symbol
            The relative distance from the start node to the measured `point`.
        """
        if ptype.casefold() == "geometric":
            zentry = "z"
        elif ptype.casefold() == "optical":
            zentry = "z_optical"
        else:
            raise ValueError(
                f"Unrecognised value for ptype: {ptype}. This must be either "
                "'geometric' or 'optical'."
            )

        if isinstance(point, OpticalNode):
            if point not in self.node_info:
                raise KeyError(
                    f"No optical node of name {point.full_name} in the solution."
                )
            else:
                z = self.node_info[point][zentry]
        elif isinstance(point, Port):
            ientry = self.node_info.get(point.i, {})
            oentry = self.node_info.get(point.o, {})

            if not ientry and not oentry:
                raise KeyError(f"No port of name {point.full_name} in the solution.")

            z1 = ientry.get(zentry, None)
            z2 = oentry.get(zentry, None)

            if z1 is not None and z2 is not None:
                if z1 != z2:
                    raise RuntimeError(
                        f"Bug encountered! Positions of nodes {point.i.full_name} "
                        f"and {point.o.full_name} not equal ({z1} != {z2})"
                    )
                z = z1
            else:
                z = z1 or z2
        elif isinstance(point, Connector) or isinstance(point, str):
            point = self.__get_component(point)
            z = None
            for node in point.optical_nodes:
                if node in self.node_info:
                    z = self.node_info[node][zentry]
                    break

            if z is None:
                raise KeyError(f"No component of name {point.name} in the solution.")
        else:
            raise TypeError("Unrecognised type for argument point.")

        return z

    def make_table(self, tablefmt="fancy_grid", subs=None):
        """Construct a table showing the beam properties at each node.

        Parameters
        ----------
        tablefmt : str, optional; default: "fancy_grid"
            The plain-text table format to pass to tabulate.

        subs : dict, optional
            A dictionary of model parameter to value substitutions
            to pass to the ``eval`` methods of symbolic expressions.

            If this solution object is not symbolic then this argument
            is ignored.

        Returns
        -------
        table : str
            The table as a string.
        """
        from tabulate import tabulate

        if not self.symbolic and subs is not None:
            LOGGER.warning(
                "Ignoring subs=%s kwarg as PropagationSolution is non-symbolic.", subs,
            )

        table = []
        headers = [
            "Name",
            "z [m]",
            "w0 [mm]",
            "zr [m]",
            "w [mm]",
            "RoC [m]",
            "Acc. Gouy [deg]",
            "q",
        ]

        format_q = lambda q: f"{q.real:.3f} + {q.imag:.3f}j"

        agouy = 0
        for node, info in self.node_info.items():
            name = node.full_name
            z = self._eval_z_for_display(info, subs=subs)
            q = self._eval_q_for_display(info, subs=subs)

            w0 = q.w0 / 1e-3
            zr = q.zr
            w = q.w / 1e-3
            Rc = q.Rc

            if node.is_input:  # accumulate the Gouy phase
                space_info = self.comp_info.get(node.space)
                if space_info:
                    agouy += space_info.get("acc_gouy", 0)

            if isinstance(agouy, Symbol):
                agouy_val = agouy.eval(subs=subs)
            else:
                agouy_val = agouy

            table.append([name, z, w0, zr, w, Rc, agouy_val, format_q(q)])

        return tabulate(table, headers, tablefmt=tablefmt)

    def __str__(self):
        return self.make_table()

    def print(self, tablefmt="fancy_grid", subs=None):
        """Print the propagated beam properties at each node in a table format.

        This internally calls :meth:`.PropagationSolution.make_table` and
        prints its return string.

        Parameters
        ----------
        tablefmt : str, optional; default: "fancy_grid"
            The plain-text table format to pass to tabulate.

        subs : dict, optional
            A dictionary of model parameter to value substitutions
            to pass to the ``eval`` methods of symbolic expressions.

            If this solution object is not symbolic then this argument
            is ignored.
        """
        print(self.make_table(tablefmt=tablefmt, subs=subs))

    def _eval_z_for_display(self, info, ztype="geometric", subs=None):
        if ztype.casefold() == "geometric":
            z = info["z"]
        elif ztype.casefold() == "optical":
            z = info["z_optical"]
        else:
            raise ValueError(f"Unrecognised ztype: {ztype}")

        if self.symbolic and isinstance(z, Symbol):
            if subs is not None:
                if not isinstance(subs, dict):
                    raise TypeError("Expected argument subs to be of type dict")

                if any(is_iterable(arg) for arg in subs.values()):
                    raise ValueError(
                        "Parameter substitutions for plotting / printing must "
                        "not use arrays of values."
                    )
            z = z.eval(subs=subs)

        return z

    def _eval_q_for_display(self, info, subs=None):
        q = info["q"]
        if self.symbolic and q.symbolic:
            if subs is not None:
                if not isinstance(subs, dict):
                    raise TypeError("Expected argument subs to be of type dict")

                if any(is_iterable(arg) for arg in subs.values()):
                    raise ValueError(
                        "Parameter substitutions for plotting / printing must "
                        "not use arrays of values."
                    )
            q = q.eval(subs=subs)

        return q

    def _eval_L_for_display(self, space, subs=None):
        if self.symbolic:
            if subs is not None:
                if not isinstance(subs, dict):
                    raise TypeError("Expected argument subs to be of type dict")

                if any(is_iterable(arg) for arg in subs.values()):
                    raise ValueError(
                        "Parameter substitutions for plotting / printing must "
                        "not use arrays of values."
                    )
            L = space.L.ref.eval(subs=subs)
        else:
            L = self.__frozen_space_lengths[space]

        return L

    def segment(self, node, *args, normalise_z=True, w_scale=1, npts=400, subs=None):
        """Obtain data for a segment of the beam over the space attached to the
        specified `node`.

        The expected, valid positional `args` are any combination of:

          * "beamsize",
          * "gouy",
          * "curvature",

        where all three will be used by default if none of these are given.

        Use :meth:`.PropagationSolution.all_segments` to obtain the beam data
        over all spaces of the solution.

        Parameters
        ----------
        node : :class:`.OpticalNode`
            The starting node of the segment.

        normalise_z : bool, optional; default: True
            Whether to normalise returned ``data["z"]`` array such that
            first value of this is zero.

        w_scale : scalar, optional; default: 1
            Quantity to scale beam size values by if calculating these. For
            example, specify `w_scale = 1e3` to get beam sizes in mm. By
            default the units of the beam size will be in metres.

        npts : int, optional; default: 400
            Number of points to use for computing data values.

        subs : dict, optional
            A dictionary of model parameter to value substitutions
            to pass to the ``eval`` methods of symbolic expressions.

            If this solution object is not symbolic then this argument
            is ignored.

        Returns
        -------
        zs : :class:`numpy.ndarray`
            Array of z-axis values corresponding to the position of the node
            up to the length of the attached space. If `normalise_z` is True
            then the position of `node` will be subtracted from all values, such
            that the first value in this array will be zero.

        data : dict
            Dictionary of data mapping ``args : values``, where `args` are those
            specified (see above) and `values` are the arrays of values corresponding
            to each of these args as a function of the z-axis values.
        """
        if not args:
            args = ("beamsize", "gouy", "curvature")

        if not isinstance(node, OpticalNode):
            raise NotImplementedError(
                "Obtaining beam segments from locations other than OpticalNodes "
                "is not yet supported."
            )

        if node.is_input or node is self.end_node:
            raise ValueError(
                "Cannot obtain a beam segment corresponding to "
                f"the node {node.full_name}. This node must be an "
                "output node and cannot be the end node of the solution."
            )

        info = self.node_info.get(node)
        if info is None:
            raise ValueError(f"Node {node.full_name} not present in the solution!")

        # Store the z data and each arg data array here
        data = {}

        z = self._eval_z_for_display(info, subs=subs)
        q = self._eval_q_for_display(info, subs=subs)
        # Make array of values from [0, space_length]
        L = self._eval_L_for_display(node.space, subs=subs)
        intra_zs = np.linspace(0, L, npts)

        zs = z + intra_zs
        if normalise_z:
            zs -= zs[0]

        for arg in args:
            values = getattr(q, arg)(q.z + intra_zs)

            if arg == "beamsize":
                values *= w_scale

            # Computing Gouy phase so convert to degrees and
            # subtract first value to get Gouy phase accumulated
            # over the segment rather than absolute phases
            if arg == "gouy":
                values = np.degrees(values)
                values -= values[0]

            data[arg] = values

        return zs, data

    def __adaptive_npts_per_segment(self, N, subs=None):
        dS_per_space = {}
        via_waists = {}
        end_node = self.end_node
        for (node1, node2) in pairwise(self.nodes):
            if node1.is_input or node1 is end_node:
                continue

            space = node1.space
            q1 = self._eval_q_for_display(self.node_info[node1], subs=subs)
            q2 = self._eval_q_for_display(self.node_info[node2], subs=subs)

            # Compute change in defocus over the space
            dS_per_space[space] = abs(q2.S - q1.S)
            # And determine whether we go through a waist
            via_waists[space] = np.sign(q1.S) != np.sign(q2.S)

        # Get the maximum defocus change
        max_dS = max(dS_per_space.values())
        for (space, dS), via_waist in zip(dS_per_space.items(), via_waists.values()):
            # If a waist exists in this space then want to automatically
            # increase the weighting by resetting dS to twice the maximum
            # so that resolution increases around the waist
            if via_waist:
                dS_per_space[space] = max_dS * 2

        dS_sum = sum(dS_per_space.values())
        npts_per_space = {}
        for space, dS in dS_per_space.items():
            # Now determine the number of points to use for this segment
            # by the size of the defocus change
            n = int(N * dS / dS_sum)
            # If the defocus change was zero, or small enough such that n is
            # less than two, let's just set the number of points for this
            # segment to two, this means that total number of points can
            # be a few larger than N but that's not really a problem
            if n < 2:
                n = 2

            npts_per_space[space] = n

        return npts_per_space

    def __equal_npts_per_segment(self, N, **_):
        nspaces = len(self.spaces)
        npts_per_space = {}
        for space in self.spaces:
            npts_per_space[space] = int(N / nspaces)

        return npts_per_space

    def __all_npts_per_segment(self, N, **_):
        npts_per_space = {}
        for space in self.spaces:
            npts_per_space[space] = N

        return npts_per_space

    def all_segments(
        self,
        *args,
        add_gouy=True,
        w_scale=1,
        npts=1000,
        resolution="adaptive",
        subs=None,
    ):
        """Construct a dictionary containing beam data for all segments of the solution.

        The expected, valid positional arguments are any combination of:

          * "beamsize",
          * "gouy",
          * "curvature",

        where all three will be used by default if no args are given.

        Use :meth:`.PropagationSolution.segment` to obtain the beam data
        over a single space of the solution.

        Parameters
        ----------
        add_gouy : bool, optional; default: True
            Whether to add the last Gouy phase from previous segment
            to all values of current segment, thereby constructing each
            "gouy" array entry as accumulated Gouy phases over all
            segments.

        w_scale : scalar, optional; default: 1
            Quantity to scale beam size values by if calculating these. For
            example, specify `w_scale = 1e3` to get beam sizes in mm. By
            default the units of the beam size will be in metres.

        npts : int, optional; default: 1000
            Number of points to use for computing data values. The actual number
            of data points used per segment depends upon the `resolution` argument.

        resolution : str, optional; default: "adaptive"
            The method of segment resolution setting to use. This can be one
            of three arguments:

              * "adaptive": Sets the number of points per segment in such a way as
                to attempt to increase the resolution near the waist. Each segment
                will have a number of points allocated to it accordingly, with the
                total number of points across all segments then approximately equal
                to `npts`.
              * "equal": Allocates an equal number of points to each segment, i.e.
                each segment has ``int(npts / len(self.spaces))`` points.
              * "all": Gives ``npts`` to all segments, such that the total number of
                data points across all segments is ``len(self.spaces) * npts``.

        subs : dict, optional
            A dictionary of model parameter to value substitutions
            to pass to the ``eval`` methods of symbolic expressions.

            If this solution object is not symbolic then this argument
            is ignored.

        Returns
        -------
        data : dict
            Dictionary of data mapping ``space : zs, segdata``, where `space` is
            each space (i.e. segment) in the solution path, `zs` are the z-axis
            values and `segdata` is the dict of data values for the targeted beam
            properties over the space.
        """
        data = {}

        if not args:
            args = ("beamsize", "gouy", "curvature")

        resolution_map = {
            "adaptive": self.__adaptive_npts_per_segment,
            "equal": self.__equal_npts_per_segment,
            "all": self.__all_npts_per_segment,
        }
        resolution_method = resolution_map.get(resolution.casefold())
        if resolution_method is None:
            raise ValueError(
                f"Unexpected value for resolution argument: {resolution}. Expected "
                "one of: 'adaptive', 'equal' or 'all'."
            )

        npts_per_space = resolution_method(npts, subs=subs)

        prev_gouy = 0
        end_node = self.end_node
        for node in self.nodes:
            if node.is_input or node is end_node:
                continue

            zs, segdata = self.segment(
                node,
                *args,
                # Don't want to normalise each start z to zero now as
                # we want each to start from actual node position
                normalise_z=False,
                w_scale=w_scale,
                npts=npts_per_space[node.space],
                subs=subs,
            )

            # Accumulating Gouy over all segments so make adjustments
            if "gouy" in args and add_gouy:
                gouys = segdata["gouy"]
                # Add on the last value of Gouy from prev segment so
                # that phase is accumulated over full path
                gouys += prev_gouy
                prev_gouy = gouys[-1]

            data[node.space] = zs, segdata

        return data

    def plot_beamsizes(self, **kwargs):
        """Plot the beam sizes over the propagated path.

        This is just a convenience wrapper which is identical to calling
        :meth:`PropagationSolution.plot` with ``"beamsize"`` as the arg.

        Returns
        -------
        fig : Figure
            Handle to the figure.

        ax : Axis
            Handle to the axis.
        """
        return self.plot("beamsize", **kwargs)

    def plot_acc_gouy(self, **kwargs):
        """Plot the accumulated Gouy phases over the propagated path.

        This is just a convenience wrapper which is identical to calling
        :meth:`PropagationSolution.plot` with ``"gouy"`` as the arg.

        Returns
        -------
        fig : Figure
            Handle to the figure.

        ax : Axis
            Handle to the axis.
        """
        return self.plot("gouy", **kwargs)

    def plot_curvatures(self, **kwargs):
        """Plot the wavefront curvatures over the propagated path.

        This is just a convenience wrapper which is identical to calling
        :meth:`PropagationSolution.plot` with ``"curvature"`` as the arg.

        Returns
        -------
        fig : Figure
            Handle to the figure.

        ax : Axis
            Handle to the axis.
        """
        return self.plot("curvature", **kwargs)

    def plot(
        self,
        *args,
        filename=None,
        show=True,
        ignore=None,
        name_xoffsets=None,
        name_yoffsets=None,
        ylims=None,
        npts=1000,
        resolution="adaptive",
        subs=None,
    ):
        """Plot any combination of the beam sizes, accumulated Gouy phases and / or
        wavefront curvatures over the propagated path.

        The expected, valid positional arguments are any combination of:
            - "beamsize",
            - "gouy",
            - "curvature",

        or "all" to plot all of the above.

        If no positional args are given then the beamsize (first axis) and accumulated
        Gouy phase (second axis) will be plotted by default.

        The locations of each component will be marked on the figure, unless the
        component is in `ignore` or the component has "AR" or "HR" in its name.

        Parameters
        ----------
        filename : str or file-like, optional
            Name of a file or existing file object to save the figure to.

        show : bool, optional; default: True
            Whether to show the figure.

        ignore : component, sequence of, optional
            A component or sequence of components to ignore when
            making markers.

        name_xoffsets : dict, optional
            Dictionary of component names to x-axis offsets for shifting where the
            component name text is placed. The offset value is interpreted in terms
            of data co-ordinates.

        name_yoffsets : dict, optional
            Dictionary of component names to y-axis offsets for shifting where the
            component name text is placed. The offset value is interpreted in terms
            of data co-ordinates.

        ylims : dict, optional
            Dictionary of target names (i.e. "beamsize", "gouy" or "curvature") to
            manual axis y-limits.

        npts : int, optional; default: 1000
            See equivalent argument in :meth:`.PropagationSolution.all_segments`.

        resolution : str, optional; default: "adaptive"
            See equivalent argument in :meth:`.PropagationSolution.all_segments`.

        subs : dict, optional
            A dictionary of model parameter to value substitutions
            to pass to the ``eval`` methods of symbolic expressions.

            If this solution object is not symbolic then this argument
            is ignored.

        Returns
        -------
        fig : Figure
            Handle to the figure.

        axs : axes
            The axis handles.
        """
        import matplotlib.pyplot as plt

        if not args:
            args = ("beamsize", "gouy")
        if "all" in args:
            args = ("beamsize", "gouy", "curvature")

        valid_args = ("beamsize", "gouy", "curvature")
        if any(arg not in valid_args for arg in args):
            raise ValueError(
                "Invalid target argument in args, expected any "
                f"combination of {valid_args} or 'all'"
            )

        if not self.symbolic and subs is not None:
            LOGGER.warning(
                "Ignoring subs=%s kwarg as PropagationSolution is non-symbolic.", subs,
            )

        N = len(args)
        fig, axs = plt.subplots(N, 1, sharex=True)
        if N == 1:
            axs = [axs]

        maximums = {k: 0 for k in args}
        minimums = {k: float("inf") for k in args}
        data = self.all_segments(
            *args, w_scale=1e3, npts=npts, resolution=resolution, subs=subs
        )
        for zs, segd in data.values():
            for i, arg in enumerate(args):
                y = segd[arg]
                if arg == "beamsize":
                    axs[i].fill_between(zs, y, -y, color="r")
                else:
                    axs[i].plot(zs, y)

                maximums[arg] = max(maximums[arg], y.max())
                minimums[arg] = min(minimums[arg], y.min())

        if ignore is None:
            ignore = []
        if not is_iterable(ignore):
            ignore = [ignore]

        if name_xoffsets is None:
            name_xoffsets = {}
        if name_yoffsets is None:
            name_yoffsets = {}

        for node, info in self.node_info.items():
            if node.is_input:
                z = self._eval_z_for_display(info, subs=subs)

                comp = node.component
                if "AR" not in comp.name and comp not in ignore:
                    name = comp.name
                    x_offset = name_xoffsets.get(name, 0)
                    y_offset = name_yoffsets.get(name, 0)
                    # display the name in a nicer way
                    name = name.replace("_", "\n")
                    n_newlines = name.count("\n")

                    for i, arg in enumerate(args):
                        if not i:
                            axs[i].axvline(
                                z, 0.12 + 0.1 * n_newlines, color="k", linestyle="--"
                            )

                            if arg == "beamsize":
                                ytext_pos = -1 * maximums[arg]
                            else:
                                ytext_pos = minimums[arg]

                            ytext_pos += y_offset
                            axs[i].text(
                                z + x_offset, ytext_pos, name, ha="center", va="center"
                            )
                        else:
                            axs[i].axvline(z, color="k", linestyle="--")

        for ax in axs:
            ax.set_xlim(0, None)

        axs[-1].set_xlabel("Distance [m]")

        ylabel_mappings = {
            "beamsize": "Beam size [mm]",
            "gouy": "Gouy phase\naccumulation [deg]",
            "curvature": "Wavefront curvature [1/m]",
        }
        if ylims is None:
            ylims = {}
        for i, arg in enumerate(args):
            if arg != "beamsize":
                ylim = ylims.get(arg, None)
                if ylim is None:
                    axs[i].set_ylim(0 if arg == "gouy" else None, maximums[arg])
                else:
                    axs[i].set_ylim(ylim[0], ylim[1])

            ylabel = ylabel_mappings.get(arg, arg)
            axs[i].set_ylabel(ylabel)

        if filename is not None:
            fig.savefig(filename)
        if show:
            plt.show()

        if N == 1:
            return fig, axs[0]

        return fig, axs

    def animate_beamsizes(self, subs, **kwargs):
        """Animate the beam sizes over the propagated path.

        This is just a convenience wrapper which is identical to calling
        :meth:`PropagationSolution.animate` with ``"beamsize"`` as the arg.

        Returns
        -------
        fig : Figure
            Handle to the figure.

        ax : Axis
            Handle to the axis.

        an : FuncAnimation
            Handle to the animation.
        """
        return self.animate(subs, "beamsize", **kwargs)

    def animate_acc_gouy(self, subs, **kwargs):
        """Animate the accumulated Gouy phases over the propagated path.

        This is just a convenience wrapper which is identical to calling
        :meth:`PropagationSolution.animate` with ``"gouy"`` as the arg.

        Returns
        -------
        fig : Figure
            Handle to the figure.

        ax : Axis
            Handle to the axis.

        an : FuncAnimation
            Handle to the animation.
        """
        return self.animate(subs, "gouy", **kwargs)

    def animate_curvatures(self, subs, **kwargs):
        """Animate the wavefront curvatures over the propagated path.

        This is just a convenience wrapper which is identical to calling
        :meth:`PropagationSolution.animate` with ``"curvature"`` as the arg.

        Returns
        -------
        fig : Figure
            Handle to the figure.

        ax : Axis
            Handle to the axis.

        an : FuncAnimation
            Handle to the animation.
        """
        return self.animate(subs, "curvature", **kwargs)

    def animate(
        self,
        subs,
        *args,
        filename=None,
        show=True,
        ignore=None,
        name_xoffsets=None,
        name_yoffsets=None,
        ylims=None,
        npts=200,
        blit=True,
        interval=200,
    ):
        """Animate any combination of the beam sizes, accumulated Gouy phases and / or
        wavefront curvatures over the propagated path using the substitution parameters
        in `subs`.

        The expected, valid positional arguments (i.e. `*args`) are any combination of:
            - "beamsize",
            - "gouy",
            - "curvature",

        or "all" to animate all of the above.

        If no positional args are given then the beamsize (first axis) and accumulated
        Gouy phase (second axis) will be animated by default.

        At least one model parameter substitution in `subs` must be
        array-like - this will then be the animation axis. If more than
        one are array-like then each array must be the same size - the
        substitutions will then be carried out simulatenously. Any scalar
        value entry in `subs` will be applied before the animation axis.

        Parameters
        ----------
        subs : dict
            Dictionary of model parameter substitutions. At least one entry
            must be array-like such than animation can be performed over this
            axis.

            If multiple substitutions are arrays then they must all be the same
            size.

        filename : str, optional
            Name of a file to save the animation to.

        show : bool, optional; default: True
            Whether to show the resulting animation.

        ignore : component, sequence of, optional
            A component or sequence of components to ignore when
            making markers.

        name_xoffsets : dict, optional
            Dictionary of component names to x-axis offsets for shifting where the
            component name text is placed. The offset value is interpreted in terms
            of data co-ordinates.

        name_yoffsets : dict, optional
            Dictionary of component names to y-axis offsets for shifting where the
            component name text is placed. The offset value is interpreted in terms
            of data co-ordinates.

        ylims : dict, optional
            Dictionary of target names (i.e. "beamsize", "gouy" or "curvature") to
            manual axis y-limits.

        npts : int, optional; default: 200
            Number of points to use for computing beam sizes and
            Gouy phases over spaces.

        blit : bool, optional; default: True
            Whether blitting is used to optimize drawing.

        interval : int, optional; default: 200
            Delay between frames in milliseconds.

        Returns
        -------
        fig : Figure
            Handle to the figure.

        axs : axes
            The axis handles.

        an : FuncAnimation
            Handle to the animation.
        """
        import matplotlib.pyplot as plt
        import matplotlib.animation as animation

        if not self.symbolic:
            raise RuntimeError(
                "Animation is only applicable to symbolic PropagationSolution objects"
            )

        if not args:
            args = ("beamsize", "gouy")
        if "all" in args:
            args = ("beamsize", "gouy", "curvature")

        valid_args = ("beamsize", "gouy", "curvature")
        if any(arg not in valid_args for arg in args):
            raise ValueError(
                "Invalid target argument in args, expected any "
                f"combination of {valid_args} or 'all'"
            )

        anim_params = dict(filter(lambda x: is_iterable(x[1]), subs.items()))
        if not anim_params:
            raise ValueError("Expected at least one array-like substitution in subs.")
        N = len(list(anim_params.values())[0])
        if any(len(arr) != N for arr in anim_params.values()):
            raise ValueError("Lengths of all array-like substitutions must be equal.")

        # Dictionary of singular value substitutions to use at each animation point
        isubs = {}

        static_params = dict(filter(lambda x: not is_iterable(x[1]), subs.items()))
        # Set fixed parameter substitutions
        for param, value in static_params.items():
            isubs[param] = value

        if ignore is None:
            ignore = []
        if not is_iterable(ignore):
            ignore = [ignore]

        # Data dictionaries storing arrays of propagated distances, beamsizes
        # accumulated Gouy phases and component positions at each node
        # for each animation point
        node_xdata = {}
        node_ydata = {k: {} for k in args}

        comp_positions = {}
        for node in self.node_info:
            if not node.is_input:
                node_xdata[node] = np.zeros((N, npts))
                for arg in args:
                    node_ydata[arg][node] = np.zeros((N, npts))
            else:
                comp = node.component
                if "AR" not in comp.name and comp not in ignore:
                    comp_positions[node] = np.ma.zeros(N)

        maximums = {k: 0 for k in args}
        minimums = {k: float("inf") for k in args}
        end_node = self.end_node
        for i in range(N):
            # Update animation parameters
            for param, array in anim_params.items():
                isubs[param] = array[i]

            # TODO (sjr) Replace this loop with an all_segments call, will require
            #            some careful rejigging of code in this method so I'll leave
            #            this for a later date
            prev_acc_gouy = 0
            for node, info in self.node_info.items():
                if not node.is_input and node != end_node:
                    z = self._eval_z_for_display(info, subs=isubs)
                    q = self._eval_q_for_display(info, subs=isubs)
                    # Make array of values from [0, space_length]
                    L = self._eval_L_for_display(node.space, subs=isubs)
                    intra_zs = np.linspace(0, L, npts)

                    node_xdata[node][i][:] = z + intra_zs
                    for arg in args:
                        data = getattr(q, arg)(q.z + intra_zs)
                        if arg == "beamsize":
                            data /= 1e-3  # scale to mm
                        elif arg == "gouy":
                            # Plotting accumulated Gouy so need to modify data as follows:
                            #  - convert to degrees
                            #  - subtract the Gouy phase directly at the output node to get
                            #    Gouy phase accumulated over this space
                            #  - then add previous accumulated Gouy to keep cumulative over full path
                            data = np.degrees(data)
                            data -= data[0]
                            data += prev_acc_gouy
                            prev_acc_gouy = data[-1]

                        maximums[arg] = max(maximums[arg], data.max())
                        minimums[arg] = min(minimums[arg], data.min())
                        node_ydata[arg][node][i][:] = data
                else:
                    if node in comp_positions:
                        comp_positions[node][i] = self._eval_z_for_display(
                            info, subs=isubs
                        )

        Np = len(args)
        fig, axs = plt.subplots(Np, 1, sharex=True)
        if Np == 1:
            axs = [axs]

        # Lists of the various artists to animate
        collections = []  # beam size fill shapes
        lines = {k: [] for k in args if k != "beamsize"}
        comp_pos_lines = []  # component position lines
        comp_name_texts = []  # names of components
        for i, arg in enumerate(args):
            for xdata, ydata in zip(node_xdata.values(), node_ydata[arg].values()):
                if arg == "beamsize":
                    collections.append(
                        axs[i].fill_between(xdata[0], ydata[0], -ydata[0], color="r")
                    )
                else:
                    (line,) = axs[i].plot([], [])
                    lines[arg].append(line)

        if name_xoffsets is None:
            name_xoffsets = {}
        if name_yoffsets is None:
            name_yoffsets = {}

        comp_name_offsets = []
        for node, zs in comp_positions.items():
            name = node.component.name
            x_offset = name_xoffsets.get(name, 0)
            y_offset = name_yoffsets.get(name, 0)
            name = name.replace("_", "\n")
            n_newlines = name.count("\n")

            for i, arg in enumerate(args):
                z0 = zs[0]
                if not i:
                    vline = axs[i].axvline(
                        z0, 0.12 + 0.1 * n_newlines, color="k", linestyle="--"
                    )

                    if arg == "beamsize":
                        ytext_pos = -1 * maximums[arg]
                    else:
                        ytext_pos = minimums[arg]

                    ytext_pos += y_offset
                    ct = axs[i].text(
                        z0 + x_offset, ytext_pos, name, ha="center", va="center"
                    )
                    comp_name_texts.append(ct)
                    comp_name_offsets.append(x_offset)
                else:
                    vline = axs[i].axvline(z0, color="k", linestyle="--")

                comp_pos_lines.append(vline)

        anim_txt_former = lambda k: "\n".join(
            f"{p.component.name} {p.name} = {v[k]:.2f} {p.units}"
            for p, v in anim_params.items()
        )
        anim_text = axs[-1].text(
            0.15, 0.9, "", ha="center", va="center", transform=axs[-1].transAxes,
        )

        axs[-1].set_xlabel("Distance [m]")

        ylabel_mappings = {
            "beamsize": "Beam size [mm]",
            "gouy": "Gouy phase\naccumulation [deg]",
            "curvature": "Wavefront curvature [1/m]",
        }
        if ylims is None:
            ylims = {}
        for i, arg in enumerate(args):
            if arg != "beamsize":
                ylim = ylims.get(arg, None)
                if ylim is None:
                    axs[i].set_ylim(0 if arg == "gouy" else None, maximums[arg])
                else:
                    axs[i].set_ylim(ylim[0], ylim[1])

            ylabel = ylabel_mappings.get(arg, arg)
            axs[i].set_ylabel(ylabel)

        # Set maximum propagated distance
        max_x = 0
        for xdata in node_xdata.values():
            max_x = max(max_x, xdata.max())
        for ax in axs:
            ax.set_xlim(0, max_x)

        try:
            w_ax_index = args.index("beamsize")
        except ValueError:
            w_ax_index = -1

        all_artists = collections + comp_pos_lines + comp_name_texts + [anim_text]
        for arg in args:
            if arg == "beamsize":
                continue
            all_artists += lines[arg]

        # Need to create a dummy fig and ax for making a fill_between collection
        # in each segment -> the actual collections then get modified using the
        # vertices of this dummy collection during the animation
        dummy_fig, dummy_ax = plt.subplots()

        def animate(frame_index):
            if w_ax_index != -1:
                for coll, xdata, ydata_w in zip(
                    collections, node_xdata.values(), node_ydata["beamsize"].values(),
                ):
                    dummy_coll = dummy_ax.fill_between(
                        xdata[frame_index],
                        ydata_w[frame_index],
                        -ydata_w[frame_index],
                        color="r",
                    )
                    dummy_path = dummy_coll.get_paths()[0]
                    path = coll.get_paths()[0]
                    path.vertices[:, 0] = dummy_path.vertices[:, 0]
                    path.vertices[:, 1] = dummy_path.vertices[:, 1]

            for arg in args:
                if arg == "beamsize":
                    continue

                for line, xdata, ydata in zip(
                    lines[arg], node_xdata.values(), node_ydata[arg].values(),
                ):
                    line.set_xdata(xdata[frame_index])
                    line.set_ydata(ydata[frame_index])

            for j, zs in zip(
                range(0, len(comp_pos_lines), Np), comp_positions.values()
            ):
                for i in range(Np):
                    vline = comp_pos_lines[j + i]
                    vline.set_xdata(zs[frame_index])

            for ctname_txt, offset, zs in zip(
                comp_name_texts, comp_name_offsets, comp_positions.values()
            ):
                ctname_txt.set_x(zs[frame_index] + offset)

            anim_text.set_text(anim_txt_former(frame_index))

            return all_artists

        an = animation.FuncAnimation(
            fig, animate, frames=N, blit=blit, interval=interval
        )
        # Get rid of the dummy figure now so that it doesn't get shown / saved
        plt.close(dummy_fig)

        if filename is not None:
            an.save(filename)

        if show:
            plt.show()

        if Np == 1:
            return fig, axs[0], an

        return fig, axs, an


class AstigmaticPropagationSolution(BaseSolution):
    """Solution representation of a call to
    :func:`finesse.tracing.tools.propagate_beam_astig`.

    Internally this stores two :class:`.PropagationSolution` instances which are used to
    access the per-plane beam parameters. These propagation solutions can be accessed
    via :attr:`.AstigmaticPropagationSolution.ps_x` and
    :attr:`.AstigmaticPropagationSolution.ps_y` for the tangential and sagittal planes,
    respectively.
    """

    def __init__(self, name, ps_x: PropagationSolution, ps_y: PropagationSolution):
        super().__init__(name)
        self.__ps_x = ps_x
        self.__ps_y = ps_y
        self.__symbolic = ps_x.symbolic

        self.empty = False

    @property
    def symbolic(self):
        """Whether the astigmatism solution is symbolic."""
        return self.__symbolic

    @property
    def ps_x(self):
        """The internal :class:`.PropagationSolution` for the tangential plane."""
        return self.__ps_x

    @property
    def ps_y(self):
        """The internal :class:`.PropagationSolution` for the sagittal plane."""
        return self.__ps_y

    @property
    def start_node(self):
        """The starting node of the propagation."""
        return self.ps_x.start_node

    @property
    def end_node(self):
        """The final node of the propagation."""
        return self.ps_x.end_node

    @property
    def nodes(self):
        """A list of all the nodes traversed, in order."""
        return self.ps_x.nodes

    @property
    def ports(self):
        """A list of all the ports traversed, in order."""
        return self.ps_x.ports

    @property
    def spaces(self):
        """A list of all spaces traversed, in order."""
        return self.ps_x.spaces

    @property
    def components(self):
        """A list of all components (excluding spaces) traversed, in order."""
        return self.ps_x.components

    @property
    def path_length(self):
        """The geometric path length of the traversed path."""
        return self.ps_x.path_length

    @property
    def overlaps(self):
        """A dict of nodes to the qx, qy overlaps at these nodes."""
        return {node: self.overlap(node) for node in self.nodes}

    def qx(self, at, which=None):
        """Tangential beam parameter at a given location of the path.

        See :meth:`.PropagationSolution.q` for parameters and return object
        descriptions.
        """
        return self.ps_x.q(at, which)

    def qy(self, at, which=None):
        """Sagittal beam parameter at a given location of the path.

        See :meth:`.PropagationSolution.q` for parameters and return object
        descriptions.
        """
        return self.ps_y.q(at, which)

    def overlap(self, at, which=None):
        """Overlap between tangential and sagittal beam parameters at a given location
        of the path.

        See :meth:`.PropagationSolution.q` for parameters description.

        Returns
        -------
        O : float or :class:`.Operation`
            If `at` is an :class:`.OpticalNode` instance, or a :class:`.Connector`
            object *and* `which` was specified, then the plane overlap corresponding to
            this node is returned.

        Os : dict
            Otherwise, a dictionary of the input and output node mappings to their
            respective plane overlaps is returned.
        """
        qxs = self.qx(at, which)
        qys = self.qy(at, which)

        if isinstance(at, OpticalNode) or (isinstance(at, Connector) and which):
            return BeamParam.overlap(qxs, qys)

        return {
            node: BeamParam.overlap(qx, qy)
            for node, qx, qy in zip(qxs.keys(), qxs.values(), qys.values())
        }

    def plot(
        self,
        *args,
        filename=None,
        show=True,
        ignore=None,
        name_xoffsets=None,
        name_yoffsets=None,
        ylims=None,
        npts=1000,
        resolution="equal",
        subs=None,
    ):
        """Plot any combination of the beam sizes, accumulated Gouy phases and / or
        wavefront curvatures over the propagated path, showing the values for both
        planes.

        The expected, valid positional arguments are any combination of:
            - "beamsize",
            - "gouy",
            - "curvature",

        or "all" to plot all of the above.

        If no positional args are given then the beamsize (first axis) and accumulated
        Gouy phase (second axis) will be plotted by default.

        .. note::

            The resulting figure will be divided into two columns, the first giving the
            absolute quantity values for both planes and the second giving the difference
            between the quantities in the two planes (tangential plane minus sagittal plane).

            For beam size plots, the tangential plane values are shown in red whilst
            sagittal are shown in blue. Whilst for any other quantity, the tangential
            plane values are solid lines and sagittal are dashed lines.

        The locations of each component will be marked on the figure, unless the
        component is in `ignore` or the component has "AR" or "HR" in its name.

        Parameters
        ----------
        filename : str or file-like, optional
            Name of a file or existing file object to save the figure to.

        show : bool, optional; default: True
            Whether to show the figure.

        ignore : component, sequence of, optional
            A component or sequence of components to ignore when
            making markers.

        name_xoffsets : dict, optional
            Dictionary of component names to x-axis offsets for shifting where the
            component name text is placed. The offset value is interpreted in terms
            of data co-ordinates.

        name_yoffsets : dict, optional
            Dictionary of component names to y-axis offsets for shifting where the
            component name text is placed. The offset value is interpreted in terms
            of data co-ordinates.

        ylims : dict, optional
            Dictionary of target names (i.e. "beamsize", "gouy" or "curvature") to
            manual axis y-limits.

        npts : int, optional; default: 1000
            See equivalent argument in :meth:`.PropagationSolution.all_segments`.

        resolution : str, optional; default: "equal"
            See equivalent argument in :meth:`.PropagationSolution.all_segments`.

        subs : dict, optional
            A dictionary of model parameter to value substitutions
            to pass to the ``eval`` methods of symbolic expressions.

            If this solution object is not symbolic then this argument
            is ignored.

        Returns
        -------
        fig : Figure
            Handle to the figure.

        axs : axes
            The axis handles.
        """
        import matplotlib.pyplot as plt

        if not args:
            args = ("beamsize", "gouy")
        if "all" in args:
            args = ("beamsize", "gouy", "curvature")

        valid_args = ("beamsize", "gouy", "curvature")
        if any(arg not in valid_args for arg in args):
            raise ValueError(
                "Invalid target argument in args, expected any "
                f"combination of {valid_args} or 'all'"
            )

        if not self.symbolic and subs is not None:
            LOGGER.warning(
                "Ignoring subs=%s kwarg as AstigmaticPropagationSolution is non-symbolic.",
                subs,
            )

        N = len(args)
        fig, axs = plt.subplots(N, 2, sharex=True)
        if N == 1:
            axs = np.array([axs])

        maximums = {k: 0 for k in args}
        diff_mins = {k: 0 for k in args}

        if resolution.casefold() == "adaptive":
            raise NotImplementedError(
                "Adaptive resolution method for astigmatic propagation solution "
                "plotting has an outstanding issue and so is not yet supported."
            )

        data_x = self.ps_x.all_segments(
            w_scale=1e3, npts=npts, resolution=resolution, subs=subs
        )
        data_y = self.ps_y.all_segments(
            w_scale=1e3, npts=npts, resolution=resolution, subs=subs
        )

        for (zs_x, segd_x), (zs_y, segd_y) in zip(data_x.values(), data_y.values()):
            for i, arg in enumerate(args):
                vx = segd_x[arg]
                vy = segd_y[arg]
                if arg == "beamsize":
                    axs[i][0].fill_between(zs_x, vx, -vx, color="r", alpha=0.5)
                    axs[i][0].fill_between(zs_y, vy, -vy, color="b", alpha=0.5)
                else:
                    axs[i][0].plot(zs_x, vx)
                    axs[i][0].plot(zs_y, vy, linestyle="--")

                # TODO (sjr) This currently won't work with adaptive resolution
                #            method as vx and vy may be different size arrays.
                #            Need to find a way to resolve this. Maybe just force
                #            each segment in data_x, data_y to have same points
                #            between the two (bit annoying to handle though).
                axs[i][1].plot(zs_x, vx - vy, color="k")

                maximums[arg] = max(maximums[arg], vx.max(), vy.max())
                diff_mins[arg] = min(diff_mins[arg], (vx - vy).min())

        if ignore is None:
            ignore = []
        if not is_iterable(ignore):
            ignore = [ignore]

        if name_xoffsets is None:
            name_xoffsets = {}
        if name_yoffsets is None:
            name_yoffsets = {}

        for node, info in self.ps_x.node_info.items():
            if node.is_input:
                z = self.ps_x._eval_z_for_display(info, subs=subs)

                comp = node.component
                if "AR" not in comp.name and comp not in ignore:
                    name = comp.name
                    x_offset = name_xoffsets.get(name, 0)
                    y_offset = name_yoffsets.get(name, 0)
                    # display the name in a nicer way
                    name = name.replace("_", "\n")
                    n_newlines = name.count("\n")

                    for i, arg in enumerate(args):
                        if not i:
                            for ax in axs[i]:
                                ax.axvline(
                                    z,
                                    0.12 + 0.1 * n_newlines,
                                    color="k",
                                    linestyle="--",
                                )

                            if arg == "beamsize":
                                ytext_pos = [-1 * maximums[arg], diff_mins[arg]]
                            else:
                                ytext_pos = [0, 0]

                            for ax, ytp in zip(axs[i], ytext_pos):
                                ax.text(
                                    z + x_offset,
                                    ytp + y_offset,
                                    name,
                                    ha="center",
                                    va="center",
                                )
                        else:
                            for ax in axs[i]:
                                ax.axvline(z, color="k", linestyle="--")

        for ax in axs.flatten():
            ax.set_xlim(0, None)

        for ax in axs[-1]:
            ax.set_xlabel("Distance [m]")

        ylabel_mappings = {
            "beamsize": "Beam size [mm]",
            "gouy": "Gouy phase\naccumulation [deg]",
            "curvature": "Wavefront curvature [1/m]",
        }
        ylabel_diff_mappings = {
            "beamsize": "$w_x - w_y$ [mm]",
            "gouy": r"$\psi_x - \psi_y$ [deg]",
            "curvature": "$S_x - S_y$ [1/m]",
        }
        if ylims is None:
            ylims = {}
        for i, arg in enumerate(args):
            if arg != "beamsize":
                ylim = ylims.get(arg, None)
                if ylim is None:
                    axs[i][0].set_ylim(0 if arg == "gouy" else None, maximums[arg])
                else:
                    axs[i][0].set_ylim(ylim[0], ylim[1])

            ylabel = ylabel_mappings.get(arg, arg)
            axs[i][0].set_ylabel(ylabel)

            dylabel = ylabel_diff_mappings.get(arg)
            axs[i][1].set_ylabel(dylabel)

        if filename is not None:
            fig.savefig(filename)
        if show:
            plt.show()

        if N == 1:
            return fig, axs[0]

        return fig, axs


class BeamTraceSolution(BaseSolution):
    """Trace solution corresponding to calls to :meth:`.Model.beam_trace`.

    Note that BeamTraceSolution objects are returned via :meth:`.Model.beam_trace`
    calls, they should never need to be created manually.

    This class provides a dict-like interface to beam trace solution data. If
    ``trace`` is an instance of this class then one can access the beam parameters
    at both planes of a node via::

        # Using the look-up key notation
        qx, qy = trace[node]

        # Or the get method
        qx, qy = trace.get(node)

    One can also access individual plane beam parameters with::

        # Tangential plane
        qx = trace[node].qx

        # Sagittal plane
        qy = trace[node].qy

    A copy of the Python dictionary which stores all the underlying ``node : (qx, qy)``
    mappings can be obtained with the :attr:`.BeamTraceSolution.data` property.

    To draw a forest representation of the trace solution data, one can simply do::

        # Prints the forest of beam parameters
        print(trace)

        # Stores the forest as a string
        trace_str = str(trace)

    This forest will be ordered by the trace order used for the associated :meth:`.Model.beam_trace`
    call which constructed this solution.
    """

    NodeData = typing.NamedTuple("NodeData", [("qx", BeamParam), ("qy", BeamParam)])

    def __init__(self, name, data, forest=None):
        super().__init__(name)
        self._data = {node: self.NodeData(qx, qy) for node, (qx, qy) in data.items()}

        self.__forest = forest
        self.empty = False  # tree drawing fill circle

        # TODO (sjr) Temporary workaround for #328, need better output colouring
        interactive = is_interactive()
        self.__start_highlight = ": \x1b[0;32;40m[" if not interactive else ": ["
        self.__end_highlight = "]\x1b[0m" if not interactive else "]"

    def items(self):
        """A view on the underlying dict items."""
        return self._data.items()

    def values(self):
        """A view on the underlying dict values."""
        return self._data.values()

    def keys(self):
        """A view on the underlying dict keys."""
        return self._data.keys()

    @property
    def data(self):
        """A copy of the underlying dictionary of beam trace solution data.

        :getter: Returns the beam trace solution data. Read-only.
        """
        return self._data.copy()

    @property
    def data_qx(self):
        """A copy of the underlying `data` dictionary but with only the tangential plane
        beam parameters selected.

        :getter: Returns a dictionary of traced nodes with corresponding
                 tangential plane beam parameters (read-only).
        """
        return {node: qx for node, (qx, _) in self._data.items()}

    @property
    def data_qy(self):
        """A copy of the underlying `data` dictionary but with only the sagittal plane
        beam parameters selected.

        :getter: Returns a dictionary of traced nodes with corresponding
                 sagittal plane beam parameters (read-only).
        """
        return {node: qy for node, (_, qy) in self._data.items()}

    def __getitem__(self, node):
        if isinstance(node, OpticalNode):
            return self._data[node]

        if isinstance(node, Port):
            in_q = self._data.get(node.i, None)
            out_q = self._data.get(node.o, None)

            if in_q is None and out_q is None:
                raise KeyError(f"Port {node} not in the trace data.")

            return {node.i: in_q, node.o: out_q}

        if isinstance(node, Connector):
            return {n: self._data[n] for n in node.optical_nodes}

        raise TypeError(
            "Expected key to be of type OpticalNode, Port or Connector "
            f"but got object of type: {type(node)}"
        )

    def get(self, node, default=None):
        """Gets the beam parameter(s) at the specified node / port.

        Note that `node` can be an instance of :class:`.OpticalNode`, resulting in
        this method returning qx and qy for that specific node, *or* it can be an
        object of type :class:`.Port` - in which case a dictionary of both the
        input and output node beam parameters are returned.

        Parameters
        ----------
        node : :class:`.OpticalNode` or :class:`.Port`
            The node or port to access.

        default : any, optional; default: None
            The value to return if `node` does not exist within the trace data.
        """
        try:
            return self[node]
        except KeyError:
            return default
        except TypeError:
            raise

    def q(self, node):
        """A convenience method for getting the non-astigmatic beam parameter at a node.

        .. warning::
            This is only intended to be used on nodes which do not exhibit astigmatism. If
            qx != qy at the node then this method will raise a ValueError.

            To get both qx and qy at a node use either::

                qx, qy = trace[node]

            or::

                qx, qy = trace.get(node)

            where ``trace`` is an instance of this class.

        Parameters
        ----------
        node : :class:`.OpticalNode`
            The node at which to obtain q.

        Returns
        -------
        q : :class:`.BeamParam`
            The beam parameter (which is the same in both planes) at the node.

        Raises
        ------
        ex : ValueError
            If the beam parameters qx != qy at the node.
        """
        qx, qy = self._data[node]
        if qx != qy:
            raise ValueError(
                f"Beam parameters qx != qy at node {node.full_name}. Use "
                "sub-script operator or BeamTraceSolution.get to obtain both "
                "qx and qy in general."
            )

        return qx

    def __str_q(self, q=None, node=None):
        if node is not None:
            qx, qy = self._data[node]
        if q is not None:
            if isinstance(q, tuple):
                qx, qy = q
            else:
                qx = qy = q

        format_q = lambda q: f"{q.real:.2f} + {q.imag:.2f}j"

        if qx != qy:
            return f"qx = {format_q(qx)}, qy = {format_q(qy)}"

        return f"q = {format_q(qx)}"

    def __draw_tree(self, tree, lpad, lines):
        branch = "├─"
        pipe = "│"
        end = "╰─"
        dash = "─"

        ltree = tree.left
        rtree = tree.right

        if ltree is not None:
            s = branch + dash
            if rtree is None:
                s = end + dash
                pad = "   "
            else:
                s = branch + dash
                pad = pipe + "   "
            lines.append(
                lpad
                + s
                + "o"
                + " "
                + ltree.node.full_name
                + self.__start_highlight
                + self.__str_q(node=ltree.node)
                + self.__end_highlight
            )
            self.__draw_tree(ltree, lpad + pad, lines)

        if rtree is not None:
            s = end + dash
            pad = "   "
            lines.append(
                lpad
                + s
                + "o"
                + " "
                + rtree.node.full_name
                + self.__start_highlight
                + self.__str_q(node=rtree.node)
                + self.__end_highlight
            )
            self.__draw_tree(rtree, lpad + pad, lines)

    def __str__(self):
        all_trees = ""

        format_g = lambda gx, gy: f"gx = {gx}, gy = {gy}" if gx != gy else f"g = {gx}"
        cav_info = (
            lambda cav: f"({self.__str_q(q=(cav.qx, cav.qy))}, {format_g(cav.gx, cav.gy)})"
        )

        nodes = list(self._data.keys())
        if not nodes:
            return all_trees

        if self.__forest is None:  # Drawing a full model beam trace
            forest = nodes[0]._model.trace_forest.forest
        else:  # Drawing a custom forest, e.g. from a single cavity trace
            forest = self.__forest

        for tree in forest:
            lines = [
                "    "
                + "o"
                + " "
                + tree.node.full_name
                + self.__start_highlight
                + self.__str_q(node=tree.node)
                + self.__end_highlight
            ]
            self.__draw_tree(tree, "", lines)
            tree_s = "\n    ".join(lines)

            dep = tree.dependency
            if tree.is_source and isinstance(dep, Cavity):
                all_trees += (
                    f"\nInternal trace of cavity: {dep.name} {cav_info(dep)}\n\n"
                )
            else:
                all_trees += f"\nDependency: {dep.name}\n\n"
            all_trees += tree_s + "\n"

        return all_trees

    def print(self):
        """Draws the trace solution as a forest of beam parameters.

        This uses the :class:`.TraceForest` structure associated with the model that
        constructed this solution, where each tree is ordered by the trace order of the
        relevant call to the beam tracing method.
        """
        print(str(self))



"""Tests ensuring that data points of a simulation scan get masked correctly when no
stable cavities nor gauss objects are present."""

import numpy as np
import pytest
from finesse import Model
from finesse.analysis import xaxis


@pytest.fixture
def fp_cavity_model():
    IFO = Model()
    IFO.parse(
        """
        l L0 P=1
        s s0 L0.p1 ITM.p1

        m ITM Rc=-2
        s sc ITM.p2 ETM.p1 L=1
        m ETM Rc=2

        # Add a detector relying on internal FP trace tree
        cp fp_q FP q
        # and another relying on an external FP trace tree
        bp in_w L0.p1.o w

        # Add a detector for the g-factor, the outputs of
        # this should never be masked
        cp fp_g FP g

        cav FP ITM.p2
        """
    )
    IFO.ITM.Rcy = IFO.ITM.Rcx.ref
    IFO.ETM.Rcy = IFO.ETM.Rcx.ref

    return IFO


def test_masked_critically_stable_fp(fp_cavity_model: Model):
    """Test data point with critically stable FP with g = 0 is masked
    for the relevant detectors."""
    IFO = fp_cavity_model

    # Scan lengths [1.9, 2.0, 2.1] -> 2 should give g = 0
    out = xaxis(IFO.elements["sc"].L, "lin", 1.9, 2.1, 2)
    qs = out["fp_q"]
    ws = out["in_w"]
    gs = out["fp_g"]

    assert np.ma.is_masked(qs)
    assert np.ma.is_masked(ws)

    assert np.all(np.ma.getmask(qs) == [False, True, False])
    assert np.all(np.ma.getmask(ws) == [False, True, False])

    assert not np.ma.is_masked(gs)

    assert gs[1] == 0


def test_masked_unstable_fp(fp_cavity_model: Model):
    """Test data points with critically stable -> unstable FP with
    g = 1, g > 1 masked for the relevant detectors."""
    IFO = fp_cavity_model

    # Scan lengths [3.9, 4.0, 4.1] -> 4 should give g = 1, 4.1 should give g > 1
    out = xaxis(IFO.elements["sc"].L, "lin", 3.9, 4.1, 2)
    qs = out["fp_q"]
    ws = out["in_w"]
    gs = out["fp_g"]

    assert np.ma.is_masked(qs)
    assert np.ma.is_masked(ws)

    assert np.all(np.ma.getmask(qs) == [False, True, True])
    assert np.all(np.ma.getmask(ws) == [False, True, True])

    assert not np.ma.is_masked(gs)

    assert gs[1] == 1


def test_allpoints_masked_unstable_fp(fp_cavity_model: Model):
    """Test data points with unstable FP for all points with g > 1 masked for the
    relevant detectors."""
    IFO = fp_cavity_model

    # Scan lengths in [5, 6] -> all should give g > 1
    out = xaxis(IFO.elements["sc"].L, "lin", 5, 6, 2)
    qs = out["fp_q"]
    ws = out["in_w"]
    gs = out["fp_g"]

    assert np.ma.is_masked(qs)
    assert np.ma.is_masked(ws)

    assert np.all(np.ma.getmask(qs) == [True, True, True])
    assert np.all(np.ma.getmask(ws) == [True, True, True])

    assert not np.ma.is_masked(gs)

    assert np.all(gs > 1)


@pytest.fixture
def fp_cavity_model_with_disabled_gauss():
    IFO = Model()
    IFO.parse(
        """
    l L0 P=1
    s s0 L0.p1 ITM.p1

    m ITM Rc=-2
    s sc ITM.p2 ETM.p1 L=1
    m ETM Rc=2

    # Add a detector relying on internal FP trace tree
    cp fp_q FP q
    # and another relying on an external FP trace tree
    bp in_w L0.p1.o w

    # Add a detector for the g-factor, the outputs of
    # this should never be masked
    cp fp_g FP g

    cav FP ITM.p2
    # Add a gauss at the laser now...
    gauss gL0 L0.p1.o w0=0.5m z=-0.5
    """
    )
    IFO.ITM.Rcy = IFO.ITM.Rcx.ref
    IFO.ETM.Rcy = IFO.ETM.Rcx.ref

    # ... but explicitly disable this gauss object for simulations
    IFO.sim_trace_config["disable"] = "gL0"

    return IFO


def test_masked_critically_stable_fp_disabled_gauss(
    fp_cavity_model_with_disabled_gauss: Model,
):
    """Test data point with critically stable FP with g = 0 is masked
    for the relevant detectors; when the additional Gauss object is disabled
    via sim_trace_config."""
    IFO = fp_cavity_model_with_disabled_gauss

    # Scan lengths [1.9, 2.0, 2.1] -> 2 should give g = 0
    out = xaxis(IFO.elements["sc"].L, "lin", 1.9, 2.1, 2)
    qs = out["fp_q"]
    ws = out["in_w"]
    gs = out["fp_g"]

    assert np.ma.is_masked(qs)
    assert np.ma.is_masked(ws)

    assert np.all(np.ma.getmask(qs) == [False, True, False])
    assert np.all(np.ma.getmask(ws) == [False, True, False])

    assert not np.ma.is_masked(gs)

    assert gs[1] == 0


def test_masked_unstable_fp_disabled_gauss(fp_cavity_model_with_disabled_gauss: Model):
    """Test data points with critically stable -> unstable FP with
    g = 1, g > 1 masked for the relevant detectors; when the additional
    Gauss object is disabled via sim_trace_config."""
    IFO = fp_cavity_model_with_disabled_gauss

    # Scan lengths [3.9, 4.0, 4.1] -> 4 should give g = 1, 4.1 should give g > 1
    out = xaxis(IFO.elements["sc"].L, "lin", 3.9, 4.1, 2)
    qs = out["fp_q"]
    ws = out["in_w"]
    gs = out["fp_g"]

    assert np.ma.is_masked(qs)
    assert np.ma.is_masked(ws)

    assert np.all(np.ma.getmask(qs) == [False, True, True])
    assert np.all(np.ma.getmask(ws) == [False, True, True])

    assert not np.ma.is_masked(gs)

    assert gs[1] == 1


def test_allpoints_masked_unstable_fp_disabled_gauss(
    fp_cavity_model_with_disabled_gauss: Model,
):
    """Test data points with unstable FP for all points with g > 1 masked for the
    relevant detectors; when the additional Gauss object is disabled via
    sim_trace_config."""
    IFO = fp_cavity_model_with_disabled_gauss

    # Scan lengths in [5, 6] -> all should give g > 1
    out = xaxis(IFO.elements["sc"].L, "lin", 5, 6, 2)
    qs = out["fp_q"]
    ws = out["in_w"]
    gs = out["fp_g"]

    assert np.ma.is_masked(qs)
    assert np.ma.is_masked(ws)

    assert np.all(np.ma.getmask(qs) == [True, True, True])
    assert np.all(np.ma.getmask(ws) == [True, True, True])

    assert not np.ma.is_masked(gs)

    assert np.all(gs > 1)



"""Testing the structure of TraceForests built from the model forest with
only changing trees and branches included. These are constructed as an
initial step when building the simulation."""

import pytest
from finesse import Model
import finesse.tracing.ctracer as ctracer
from ...util import assert_trace_trees_equal


@pytest.fixture
def cavity_model():
    IFO = Model()
    IFO.parse(
        """
    l L0 P=1
    s s_in L0.p1 BS1.p1 L=1
    bs BS1
    s s0 BS1.p3 ITM.p1 L=5

    m ITM R=0.99 T=0.01 Rc=-2090
    s CAV ITM.p2 ETM.p1 L=4k
    m ETM R=0.99 T=0.01 Rc=2090

    cav FP ITM.p2

    s sout ETM.p2 END.p1 L=2
    nothing END
    """
    )

    return IFO


##### Changing a parameter when only the Cavity object is present #####


@pytest.mark.parametrize(
    "change", ("CAV.L", "ITM.Rcx", "ETM.Rcy"),
)
def test_changing_cavity_geometry_forest(cavity_model: Model, change):
    """Check that the changing forest is the whole cavity model when setting
    a geometric parameter of the cavity to tunable."""
    IFO = cavity_model

    internal_cav_tree = ctracer.TraceTree.from_cavity(IFO.FP)
    external_tree_out = ctracer.TraceTree.from_node(
        IFO.ETM.p2.o, IFO.FP, symmetric=True, pre_node=IFO.ETM.p1.i,
    )
    external_tree_back = ctracer.TraceTree.from_node(
        IFO.ITM.p1.o, IFO.FP, symmetric=True, pre_node=IFO.ITM.p2.i,
    )
    back_branch_tree = ctracer.TraceTree.from_node(
        IFO.BS1.p1.i, IFO.FP, symmetric=True, exclude=[IFO.BS1.p3.o,],
    )

    expect_forest = [
        internal_cav_tree,
        external_tree_out,
        external_tree_back,
        back_branch_tree,
    ]

    comp, param = change.split(".")
    pchange = getattr(IFO.elements[comp], param)
    pchange.is_tunable = True
    with IFO.built() as sim:
        chng_forest = sim.trace_forest.forest

        assert len(expect_forest) == len(chng_forest)
        for tree1, tree2 in zip(expect_forest, chng_forest):
            assert_trace_trees_equal(tree1, tree2)

        # Only a single Cavity dependency so shouldn't be any
        # mismatch node couplings present
        assert len(sim.mismatched_node_couplings) == 0


@pytest.mark.parametrize(
    "change", ("ITM.phi", "ETM.phi", "ITM.R", "ETM.T", "ITM.xbeta", "ETM.ybeta"),
)
def test_changing_non_geometric_cavity_parameter_forest(cavity_model: Model, change):
    """Check that the changing forest is empty when only a non-geometric parameter
    is set to tunable."""
    IFO = cavity_model

    comp, param = change.split(".")
    pchange = getattr(IFO.elements[comp], param)
    pchange.is_tunable = True
    with IFO.built() as sim:
        chng_forest = sim.trace_forest.forest

        assert len(chng_forest) == 0
        assert len(sim.mismatched_node_couplings) == 0


def test_changing_post_cavity_space_length_forest(cavity_model: Model):
    """Check that the changing forest is just the tree from the end mirror
    forwards when length of space after cavity is set to tunable."""
    IFO = cavity_model

    external_tree_out = ctracer.TraceTree.from_node(
        IFO.ETM.p2.o, IFO.FP, symmetric=True,
    )

    expect_forest = [external_tree_out]

    pchange = IFO.elements["sout"].L
    pchange.is_tunable = True
    with IFO.built() as sim:
        chng_forest = sim.trace_forest.forest

        assert len(expect_forest) == len(chng_forest)
        for tree1, tree2 in zip(expect_forest, chng_forest):
            assert_trace_trees_equal(tree1, tree2)

        assert len(sim.mismatched_node_couplings) == 0


def test_changing_pre_cavity_space_length_forest(cavity_model: Model):
    """Check that the changing forest is just the tree from the input mirror
    backwards when length of space before cavity is set to tunable."""
    IFO = cavity_model

    external_tree_back = ctracer.TraceTree.from_node(
        IFO.ITM.p1.o, IFO.FP, symmetric=True,
    )
    back_branch_tree = ctracer.TraceTree.from_node(
        IFO.BS1.p1.i, IFO.FP, symmetric=True, exclude=[IFO.BS1.p3.o,],
    )

    expect_forest = [
        external_tree_back,
        back_branch_tree,
    ]

    pchange = IFO.elements["s0"].L
    pchange.is_tunable = True
    with IFO.built() as sim:
        chng_forest = sim.trace_forest.forest

        assert len(expect_forest) == len(chng_forest)
        for tree1, tree2 in zip(expect_forest, chng_forest):
            assert_trace_trees_equal(tree1, tree2)

        assert len(sim.mismatched_node_couplings) == 0


def test_changing_pre_bs_space_length_forest(cavity_model: Model):
    """Check that the changing forest is just the tree from the BS
    backwards when length of space before BS is set to tunable."""
    IFO = cavity_model

    external_tree_back = ctracer.TraceTree.from_node(
        IFO.BS1.p1.o, IFO.FP, symmetric=True,
    )

    expect_forest = [external_tree_back]

    pchange = IFO.elements["s_in"].L
    pchange.is_tunable = True
    with IFO.built() as sim:
        chng_forest = sim.trace_forest.forest

        assert len(expect_forest) == len(chng_forest)
        for tree1, tree2 in zip(expect_forest, chng_forest):
            assert_trace_trees_equal(tree1, tree2)

        assert len(sim.mismatched_node_couplings) == 0


##### Adding a Gauss at the laser for extra complexity #####


@pytest.mark.parametrize(
    "change", ("CAV.L", "ITM.Rcx", "ETM.Rcy"),
)
def test_changing_cavity_geometry_with_laser_gauss_forest(cavity_model: Model, change):
    """Check that the changing forest is the whole cavity model excluding gauss tree
    when setting a geometric parameter of cavity to tunable."""
    IFO = cavity_model
    IFO.parse("gauss gL0 L0.p1.o q=-1+1j")

    internal_cav_tree = ctracer.TraceTree.from_cavity(IFO.FP)
    external_tree_out = ctracer.TraceTree.from_node(
        IFO.ETM.p2.o, IFO.FP, symmetric=True, pre_node=IFO.ETM.p1.i,
    )
    external_tree_back = ctracer.TraceTree.from_node(
        IFO.ITM.p1.o,
        IFO.FP,
        symmetric=True,
        pre_node=IFO.ITM.p2.i,
        exclude=[IFO.BS1.p1.o],
    )

    expect_forest = [
        internal_cav_tree,
        external_tree_out,
        external_tree_back,
    ]

    # Mismatch node couplings should occur on transmission through BS1 in this case
    expect_mm_couplings = (
        (IFO.BS1.p1.i, IFO.BS1.p3.o),
        (IFO.BS1.p3.i, IFO.BS1.p1.o),
        (IFO.BS1.p2.i, IFO.BS1.p4.o),
        (IFO.BS1.p4.i, IFO.BS1.p2.o),
    )

    comp, param = change.split(".")
    pchange = getattr(IFO.elements[comp], param)
    pchange.is_tunable = True
    with IFO.built() as sim:
        chng_forest = sim.trace_forest.forest

        assert len(expect_forest) == len(chng_forest)
        for tree1, tree2 in zip(expect_forest, chng_forest):
            assert_trace_trees_equal(tree1, tree2)

        assert len(sim.mismatched_node_couplings) == len(expect_mm_couplings)
        assert set(sim.mismatched_node_couplings) == set(expect_mm_couplings)


def test_changing_pre_bs_space_length_with_laser_gauss_forest(cavity_model: Model):
    """Check that the changing forest is just the tree from the BS
    forwards (from gauss node) when setting pre-BS space length tunable."""
    IFO = cavity_model
    IFO.parse("gauss gL0 L0.p1.o q=-1+1j")

    gauss_tree_forward = ctracer.TraceTree.from_node(
        IFO.L0.p1.o, IFO.gL0, symmetric=True, is_gauss=True, exclude=[IFO.BS1.p3.o],
    )

    expect_forest = [gauss_tree_forward]

    # Mismatch node couplings should occur on transmission through BS1 in this case
    expect_mm_couplings = (
        (IFO.BS1.p1.i, IFO.BS1.p3.o),
        (IFO.BS1.p3.i, IFO.BS1.p1.o),
        (IFO.BS1.p2.i, IFO.BS1.p4.o),
        (IFO.BS1.p4.i, IFO.BS1.p2.o),
    )

    pchange = IFO.elements["s_in"].L
    pchange.is_tunable = True
    with IFO.built() as sim:
        chng_forest = sim.trace_forest.forest

        assert len(expect_forest) == len(chng_forest)
        for tree1, tree2 in zip(expect_forest, chng_forest):
            assert_trace_trees_equal(tree1, tree2)

        assert len(sim.mismatched_node_couplings) == len(expect_mm_couplings)
        assert set(sim.mismatched_node_couplings) == set(expect_mm_couplings)


@pytest.mark.parametrize(
    "change", ("gL0.w0x", "gL0.w0y", "gL0.zx", "gL0.zy"),
)
def test_changing_laser_gauss_forest(cavity_model: Model, change):
    """Check that the changing forest is just the tree from the laser
    forwards when setting a Gauss parameter tunable."""
    IFO = cavity_model
    IFO.parse("gauss gL0 L0.p1.o q=-1+1j")

    gauss_tree_forward = ctracer.TraceTree.from_node(
        IFO.L0.p1.o, IFO.gL0, symmetric=True, is_gauss=True, exclude=[IFO.BS1.p3.o],
    )

    expect_forest = [gauss_tree_forward]

    # Mismatch node couplings should occur on transmission through BS1 in this case
    expect_mm_couplings = (
        (IFO.BS1.p1.i, IFO.BS1.p3.o),
        (IFO.BS1.p3.i, IFO.BS1.p1.o),
        (IFO.BS1.p2.i, IFO.BS1.p4.o),
        (IFO.BS1.p4.i, IFO.BS1.p2.o),
    )

    comp, param = change.split(".")
    pchange = getattr(IFO.elements[comp], param)
    pchange.is_tunable = True
    with IFO.built() as sim:
        chng_forest = sim.trace_forest.forest

        assert len(expect_forest) == len(chng_forest)
        for tree1, tree2 in zip(expect_forest, chng_forest):
            assert_trace_trees_equal(tree1, tree2)

        assert len(sim.mismatched_node_couplings) == len(expect_mm_couplings)
        assert set(sim.mismatched_node_couplings) == set(expect_mm_couplings)


def test_changing_bs_roc_with_cavity_priority_forest(cavity_model: Model):
    """Check that the changing forest is the trees coming from BS1 when
    setting BS1.Rcx tunable and cavity has highest trace priority."""
    IFO = cavity_model
    IFO.parse("gauss gL0 L0.p1.o q=-1+1j priority=-1")

    to_as_port_tree = ctracer.TraceTree.from_node(
        IFO.BS1.p3.i, IFO.FP, symmetric=True, exclude=[IFO.BS1.p1.o],
    )
    to_y_port_tree = ctracer.TraceTree.from_node(
        IFO.BS1.p1.i, IFO.gL0, symmetric=True, is_gauss=True, exclude=[IFO.BS1.p3.o],
    )

    expect_forest = [to_as_port_tree, to_y_port_tree]

    # Mismatch node couplings should occur on transmission through BS1 in this case
    expect_mm_couplings = (
        (IFO.BS1.p1.i, IFO.BS1.p3.o),
        (IFO.BS1.p3.i, IFO.BS1.p1.o),
        (IFO.BS1.p2.i, IFO.BS1.p4.o),
        (IFO.BS1.p4.i, IFO.BS1.p2.o),
    )

    pchange = IFO.elements["BS1"].Rcx
    pchange.is_tunable = True
    with IFO.built() as sim:
        chng_forest = sim.trace_forest.forest

        assert len(expect_forest) == len(chng_forest)
        for tree1, tree2 in zip(expect_forest, chng_forest):
            assert_trace_trees_equal(tree1, tree2)

        assert len(sim.mismatched_node_couplings) == len(expect_mm_couplings)
        assert set(sim.mismatched_node_couplings) == set(expect_mm_couplings)


def test_changing_bs_roc_with_gauss_priority_forest(cavity_model: Model):
    """Check that the changing forest is the trees coming from BS1 when
    setting BS1.Rcx tunable and gauss has highest trace priority."""
    IFO = cavity_model
    IFO.parse("gauss gL0 L0.p1.o q=-1+1j priority=1")

    from_bs_tree_forward = ctracer.TraceTree.from_node(
        IFO.BS1.p1.i, IFO.gL0, symmetric=True, is_gauss=True, exclude=[IFO.ITM.p2.i],
    )
    branch_tree = ctracer.TraceTree.from_node(
        IFO.BS1.p3.i, IFO.gL0, symmetric=True, is_gauss=True, exclude=[IFO.BS1.p1.o],
    )

    expect_forest = [
        from_bs_tree_forward,
        branch_tree,
    ]

    # Mismatch node couplings should occur on transmission through ITM in this case
    # and also for front surface reflection due to symmetric tracing behaviour
    expect_mm_couplings = (
        (IFO.ITM.p1.i, IFO.ITM.p2.o),
        (IFO.ITM.p2.i, IFO.ITM.p1.o),
        (IFO.ITM.p1.i, IFO.ITM.p1.o),
    )

    pchange = IFO.elements["BS1"].Rcx
    pchange.is_tunable = True
    with IFO.built() as sim:
        chng_forest = sim.trace_forest.forest

        assert len(expect_forest) == len(chng_forest)
        for tree1, tree2 in zip(expect_forest, chng_forest):
            assert_trace_trees_equal(tree1, tree2)

        assert len(sim.mismatched_node_couplings) == len(expect_mm_couplings)
        assert set(sim.mismatched_node_couplings) == set(expect_mm_couplings)


@pytest.mark.parametrize(
    "change", ("CAV.L", "ITM.Rcx", "ETM.Rcy"),
)
def test_changing_cavity_geometry_with_laser_gauss_disabled_forest(
    cavity_model: Model, change
):
    """Check that the changing forest is the whole cavity model when disabling
    the laser gauss and setting a geometric parameter of cavity tunable."""
    IFO = cavity_model
    IFO.parse("gauss gL0 L0.p1.o q=-1+1j")

    IFO.sim_trace_config["disable"] = "gL0"

    internal_cav_tree = ctracer.TraceTree.from_cavity(IFO.FP)
    external_tree_out = ctracer.TraceTree.from_node(
        IFO.ETM.p2.o, IFO.FP, symmetric=True, pre_node=IFO.ETM.p1.i,
    )
    external_tree_back = ctracer.TraceTree.from_node(
        IFO.ITM.p1.o, IFO.FP, symmetric=True, pre_node=IFO.ITM.p2.i,
    )
    back_branch_tree = ctracer.TraceTree.from_node(
        IFO.BS1.p1.i, IFO.FP, symmetric=True, exclude=[IFO.BS1.p3.o,],
    )

    expect_forest = [
        internal_cav_tree,
        external_tree_out,
        external_tree_back,
        back_branch_tree,
    ]

    comp, param = change.split(".")
    pchange = getattr(IFO.elements[comp], param)
    pchange.is_tunable = True
    with IFO.built() as sim:
        chng_forest = sim.trace_forest.forest

        assert len(expect_forest) == len(chng_forest)
        for tree1, tree2 in zip(expect_forest, chng_forest):
            assert_trace_trees_equal(tree1, tree2)

        assert len(sim.mismatched_node_couplings) == 0


def test_changing_itm_roc_with_laser_gauss_forest(cavity_model: Model):
    """Check that the changing forest is the cavity forwards with laser gauss
    set as highest priority. Also checks that ITM.p1.i -> ITM.p1.o is in the
    mismatched node couplings as a special case (as these nodes aren't in
    changing forest but coupling can still change when symmetrically tracing)."""
    IFO = cavity_model
    IFO.parse("gauss gL0 L0.p1.o q=-1+1j priority=1")

    internal_cav_tree = ctracer.TraceTree.from_cavity(IFO.FP)
    external_tree_out = ctracer.TraceTree.from_node(
        IFO.ETM.p2.o, IFO.FP, symmetric=True, pre_node=IFO.ETM.p1.i,
    )

    expect_forest = [
        internal_cav_tree,
        external_tree_out,
    ]

    # Mismatch node couplings should occur on transmission through ITM in this case
    # and also for front surface reflection due to symmetric tracing behaviour
    expect_mm_couplings = (
        (IFO.ITM.p1.i, IFO.ITM.p2.o),
        (IFO.ITM.p2.i, IFO.ITM.p1.o),
        (IFO.ITM.p1.i, IFO.ITM.p1.o),
    )

    IFO.ITM.Rcx.is_tunable = True
    with IFO.built() as sim:
        chng_forest = sim.trace_forest.forest

        assert len(expect_forest) == len(chng_forest)
        for tree1, tree2 in zip(expect_forest, chng_forest):
            assert_trace_trees_equal(tree1, tree2)

        assert len(sim.mismatched_node_couplings) == len(expect_mm_couplings)
        assert set(sim.mismatched_node_couplings) == set(expect_mm_couplings)


@pytest.mark.parametrize(
    "change_cav", ("gL0.w0x", "gL0.w0y", "gL0.zx", "gL0.zy"),
)
@pytest.mark.parametrize(
    "change_gauss", ("CAV.L", "ITM.Rcx", "ETM.Rcy"),
)
def test_changing_cavity_geometry_and_laser_gauss_forest(
    cavity_model: Model, change_cav, change_gauss
):
    """Check that the changing forest is the whole cavity model when setting
    a geometric parameter of the cavity and the laser gauss to tunable."""
    IFO = cavity_model
    IFO.parse("gauss gL0 L0.p1.o q=-1+1j")

    internal_cav_tree = ctracer.TraceTree.from_cavity(IFO.FP)
    external_tree_out = ctracer.TraceTree.from_node(
        IFO.ETM.p2.o, IFO.FP, symmetric=True, pre_node=IFO.ETM.p1.i,
    )
    external_tree_back = ctracer.TraceTree.from_node(
        IFO.ITM.p1.o,
        IFO.FP,
        symmetric=True,
        pre_node=IFO.ITM.p2.i,
        exclude=[IFO.BS1.p1.o],
    )
    gauss_tree_forward = ctracer.TraceTree.from_node(
        IFO.L0.p1.o, IFO.gL0, symmetric=True, is_gauss=True, exclude=[IFO.BS1.p3.o],
    )

    expect_forest = [
        internal_cav_tree,
        external_tree_out,
        external_tree_back,
        gauss_tree_forward,
    ]

    # Mismatch node couplings should occur on transmission through BS1 in this case
    expect_mm_couplings = (
        (IFO.BS1.p1.i, IFO.BS1.p3.o),
        (IFO.BS1.p3.i, IFO.BS1.p1.o),
        (IFO.BS1.p2.i, IFO.BS1.p4.o),
        (IFO.BS1.p4.i, IFO.BS1.p2.o),
    )

    for change in change_cav, change_gauss:
        comp, param = change.split(".")
        pchange = getattr(IFO.elements[comp], param)
        pchange.is_tunable = True

    with IFO.built() as sim:
        chng_forest = sim.trace_forest.forest

        assert len(expect_forest) == len(chng_forest)
        for tree1, tree2 in zip(expect_forest, chng_forest):
            assert_trace_trees_equal(tree1, tree2)

        assert len(sim.mismatched_node_couplings) == len(expect_mm_couplings)
        assert set(sim.mismatched_node_couplings) == set(expect_mm_couplings)


@pytest.mark.parametrize(
    "change", ("CAV.L", "ITM.Rcx", "ETM.Rcy"),
)
def test_changing_cavity_geometry_and_laser_gauss_reversed_priorities_forest(
    cavity_model: Model, change
):
    """Check that the changing forest is the whole cavity model when setting
    a geometric parameter of the cavity and the laser gauss to tunable."""
    IFO = cavity_model
    IFO.parse("gauss gL0 L0.p1.o q=-1+1j priority=1")

    internal_cav_tree = ctracer.TraceTree.from_cavity(IFO.FP)
    external_tree_out = ctracer.TraceTree.from_node(
        IFO.ETM.p2.o, IFO.FP, symmetric=True, pre_node=IFO.ETM.p1.i,
    )
    gauss_tree_forward = ctracer.TraceTree.from_node(
        IFO.L0.p1.o, IFO.gL0, symmetric=True, is_gauss=True, exclude=[IFO.ITM.p2.i],
    )
    branch_tree = ctracer.TraceTree.from_node(
        IFO.BS1.p3.i, IFO.gL0, symmetric=True, is_gauss=True, exclude=[IFO.BS1.p1.o],
    )

    expect_forest = [
        internal_cav_tree,
        gauss_tree_forward,
        external_tree_out,
        branch_tree,
    ]

    # Mismatch node couplings should occur on transmission through ITM in this case
    # and also for front surface reflection due to symmetric tracing behaviour
    expect_mm_couplings = (
        (IFO.ITM.p1.i, IFO.ITM.p2.o),
        (IFO.ITM.p2.i, IFO.ITM.p1.o),
        (IFO.ITM.p1.i, IFO.ITM.p1.o),
    )

    comp, param = change.split(".")
    pchange_cav = getattr(IFO.elements[comp], param)
    pchange_cav.is_tunable = True
    pchange_gauss = IFO.elements["gL0"].zx
    pchange_gauss.is_tunable = True
    with IFO.built() as sim:
        chng_forest = sim.trace_forest.forest

        assert len(expect_forest) == len(chng_forest)
        for tree1, tree2 in zip(expect_forest, chng_forest):
            assert_trace_trees_equal(tree1, tree2)

        assert len(sim.mismatched_node_couplings) == len(expect_mm_couplings)
        assert set(sim.mismatched_node_couplings) == set(expect_mm_couplings)



"""Test cases for selecting modes to include in a model."""

import pytest
import numpy as np


def test_maxtem_off(model):
    """Test HOMs when maxtem set to off."""
    model.switch_off_homs()
    assert np.all(model.homs == [[0, 0]])


def test_maxtem_zero(model):
    """Test HOMs when maxtem set to 0."""
    model.select_modes(maxtem=0)
    assert np.all(model.homs == [[0, 0]])


def test_maxtem_on(model):
    """Test HOMs when maxtem set to 2."""
    model.select_modes(maxtem=2)
    assert np.all(model.homs == [[0, 0], [0, 1], [0, 2], [1, 0], [1, 1], [2, 0]])


def test_maxtem_on_increased(model):
    """Test HOMs when maxtem set higher."""
    model.select_modes(maxtem=1)
    assert np.all(model.homs == [[0, 0], [0, 1], [1, 0]])

    model.select_modes(maxtem=2)
    assert np.all(model.homs == [[0, 0], [0, 1], [0, 2], [1, 0], [1, 1], [2, 0]])


def test_maxtem_on_decreased(model):
    """Test HOMs when maxtem set lower."""
    model.select_modes(maxtem=2)
    assert np.all(model.homs == [[0, 0], [0, 1], [0, 2], [1, 0], [1, 1], [2, 0]])

    model.select_modes(maxtem=1)
    assert np.all(model.homs == [[0, 0], [0, 1], [1, 0]])


def test_maxtem_on_to_off(model):
    """Test HOMs when maxtem set to 1 then off."""
    model.select_modes(maxtem=1)
    assert np.all(model.homs == [[0, 0], [0, 1], [1, 0]])

    model.select_modes("off")
    assert np.all(model.homs == [[0, 0]])


def test_even_modes_specify_maxtem(model):
    """Test HOMs when modes set to even."""
    model.select_modes("even", 4)

    assert np.all(model.homs == [[0, 0], [0, 2], [0, 4], [2, 0], [2, 2], [4, 0]])


def test_odd_modes_specify_maxtem(model):
    """Test HOMs when modes set to odd."""
    model.select_modes("odd", 3)

    assert np.all(model.homs == [[0, 0], [0, 1], [0, 3], [1, 0], [1, 1], [3, 0]])


def test_tangential_modes_specify_maxtem(model):
    """Test HOMs when modes set to 5 for x-direction."""
    model.select_modes("x", 5)

    assert np.all(model.homs == [[0, 0], [1, 0], [2, 0], [3, 0], [4, 0], [5, 0]])


def test_sagittal_modes_specify_maxtem(model):
    """Test HOMs when modes set to 4 for y-direction."""
    model.select_modes("y", 4)

    assert np.all(model.homs == [[0, 0], [0, 1], [0, 2], [0, 3], [0, 4]])


def test_insert_single_mode(model):
    """Test HOMs when maxtem set to 3 and modes set to x-direction, with extra mode included."""
    model.select_modes(maxtem=3)
    model.select_modes("x", 3)
    model.include_modes("11")

    assert np.all(model.homs == [[0, 0], [1, 0], [1, 1], [2, 0], [3, 0]])


def test_insert_multiple_modes(model):
    """Test HOMs when maxtem set to even 4 modes, with extra mode included."""
    model.select_modes("even", 4)
    model.include_modes(["11", "32", "50"])

    assert np.all(
        model.homs
        == [[0, 0], [0, 2], [0, 4], [1, 1], [2, 0], [2, 2], [3, 2], [4, 0], [5, 0]]
    )


def test_negative_maxtem_is_invalid(model):
    """Test that maxtem cannot be negative."""
    with pytest.raises(ValueError):
        model.select_modes(maxtem=-1)
    with pytest.raises(ValueError):
        model.select_modes(maxtem=-314)

    assert np.all(model.homs == [[0, 0]])


def test_non_integer_maxtem_is_invalid(model):
    """Test that maxtem cannot be floats."""
    with pytest.raises(ValueError):
        model.select_modes(maxtem=3.4)
    with pytest.raises(ValueError):
        model.select_modes(maxtem=-3.4)

    assert np.all(model.homs == [[0, 0]])


def test_unique_mode_indices(model):
    """Test HOMs when manually selected modes."""
    model.select_modes(["00", "11", "22", "11", "00"])

    assert np.all(model.homs == [[0, 0], [1, 1], [2, 2]])


@pytest.mark.xfail(reason="See NOTE in Model.select_modes")
def test_warns_when_enabling(caplog):
    """ Test that a warning is emitted when maxtem = 0

    Test that a warning is emitted when
    the model is automatically swicthed
    to modal, but no higher order modes
    are set.

    Also checks that the message is only
    displayed once.
    """
    import finesse
    from finesse.enums import SpatialType
    import logging

    ifo = finesse.Model()
    ifo.parse(
        f"""

    l LaserIn P=1

    s s0 LaserIn.p1 ITM.p1 L=1

    m ITM R=0.99 T=0.01 Rc=-1
    s sCAV ITM.p2 ETM.p1 L=1.5
    m ETM R=0.99 T=0.01 Rc=1
    """
    )

    assert ifo.spatial_type == SpatialType.PLANE
    assert not ifo.is_modal
    with caplog.at_level(logging.WARNING):
        ifo.parse(
            f"""
        cavity fab1 ITM.p2 via=ETM.p1.i
        """
        )
        assert "enabled with only HG00" in caplog.text

    assert ifo.spatial_type == SpatialType.MODAL
    assert ifo.is_modal

    # Check that this is only displayed once
    caplog.clear()
    with caplog.at_level(logging.WARNING):
        ifo.parse(
            f"""
        s s02 ETM.p2 ITM2.p1 L=1
        m ITM2 R=0.99 T=0.01 Rc=-1
        s sCAV2 ITM2.p2 ETM2.p1 L=1.5
        m ETM2 R=0.99 T=0.01 Rc=1
        cavity fab2 ITM2.p2 via=ETM2.p1.i
        """
        )
        assert "enabled with only HG00" not in caplog.text



"""
Wire-type objects representing electrical connections between components.
"""

import logging
import numpy as np

from finesse.components.general import Connector, borrows_nodes

LOGGER = logging.getLogger(__name__)

from finesse.components.workspace import ConnectorWorkspace
from finesse.components.node import NodeType, NodeDirection
from finesse.parameter import float_parameter


class WireWorkspace(ConnectorWorkspace):
    pass


@float_parameter("delay", "Delay", validate="_check_delay", units="s")
@borrows_nodes()
class Wire(Connector):
    """Represents a wire between two electrical components in the interferometer configuration.

    Parameters
    ----------
    name : str, optional
        Name of newly created wire.

    portA, portB : :class:`.Port`, optional
        Ports to connect.
    """

    def __init__(self, name=None, portA=None, portB=None, delay=0):
        if (portA is None) != (portB is None):
            LOGGER.warn(
                "Can't construct a wire with only one port "
                "connected, ignoring ports."
            )
            portA = None
            portB = None

        if portA is not None and portA.type != NodeType.ELECTRICAL:
            raise Exception("PortA argument is not an electrical port")
        if portB is not None and portB.type != NodeType.ELECTRICAL:
            raise Exception("PortB argument is not an electrical port")

        if name is None:
            if portA is not None and portB is not None:
                compA = portA.component.name
                compB = portB.component.name
                name = f"{compA}_{portA.name}__{compB}_{portB.name}"
            else:
                raise ValueError(
                    "Cannot create an unconnected wire without " "providing a name"
                )

        super().__init__(name)
        self.delay = delay
        self._add_to_model_namespace = False

        self.__portA = portA
        self.__portB = portB

        self._add_port("p1", NodeType.ELECTRICAL)
        self._add_port("p2", NodeType.ELECTRICAL)

        if portA is not None and portB is not None:
            self.connect(portA, portB)

    @property
    def portA(self):
        return self.__portA

    @property
    def portB(self):
        return self.__portB

    def _check_delay(self, value):
        if value < 0:
            raise ValueError("Delay of a wire must not be negative.")

        return value

    def connect(self, portA, portB):
        """
        Sets the ports of this `Wire`.

        Parameters
        ----------
        portA : :class:`.Port`, optional
            Port to connect

        portB : :class:`.Port`, optional
            Port to connect
        """
        if portA.nodes[0].direction == portB.nodes[0].direction:
            raise Exception(
                f"Can not connect two {portA.nodes[0].direction} nodes, {portA.nodes[0]} to {portB.nodes[0]}."
            )

        # From the Wire's perspective the input and output
        # nodes are swapped around for its ports
        if portA.nodes[0].direction == NodeDirection.INPUT:
            self.p1._add_node("i", None, portB.o)
            self.p2._add_node("o", None, portA.i)
        else:
            self.p1._add_node("i", None, portA.o)
            self.p2._add_node("o", None, portB.i)

        self._register_node_coupling("P1i_P2o", self.p1.i, self.p2.o)

    def _get_workspace(self, sim):
        if sim.signal:
            _, is_changing = self._eval_parameters()
            # Most wires are zero delay, so don't bother refilling them all the time
            refill = (
                (sim.model.fsig.f.is_changing or sim.signal.any_frequencies_changing)
                and self.delay.is_changing
                and self.delay != 0
            )

            ws = WireWorkspace(self, sim, refill, refill)
            ws.frequencies = sim.signal.electrical_frequencies[self.p1.i].frequencies
            # Set the fill function for this simulation
            ws.signal.add_fill_function(self.fill, refill)
            return ws
        else:
            return None

    def fill(self, ws):
        for _ in ws.frequencies:
            if ws.signal.connections.P1i_P2o_idx > -1:
                with ws.sim.signal.component_edge_fill3(
                    ws.owner_id, ws.signal.connections.P1i_P2o_idx, _.index, _.index,
                ) as mat:
                    mat[:] = np.exp(1j * ws.values.delay * _.f)



"""Memoize wrappers for the parser.

Based on https://github.com/we-like-parsers/pegen/blob/main/story5/memo.py by Guido van
Rossum.

Copyright (c) 2021 we-like-parsers

Permission is hereby granted, free of charge, to any person obtaining a copy of this
software and associated documentation files (the "Software"), to deal in the Software
without restriction, including without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
"""


def memoize(func):
    """Memoize a parsing method.

    The functon must be a method on a class deriving from Parser. The method must have
    either no arguments or a single argument that is an int or str (the latter being the
    case for expect()). It must return either None or an object that is not modified (at
    least not while we're parsing). We memoize positive and negative outcomes per input
    position. The function is expected to move the input position iff it returns a not-
    None value. The memo is structured as a dict of dict, the outer dict indexed by
    input position, the inner by function and arguments.
    """

    def memoize_wrapper(self, *args):
        pos = self.mark()
        memo = self.memos.get(pos)
        if memo is None:
            memo = self.memos[pos] = {}
        key = (func, args)
        if key in memo:
            res, endpos = memo[key]
            self.reset(endpos)
        else:
            res = func(self, *args)
            endpos = self.mark()
            if res is None:
                assert endpos == pos
            else:
                assert endpos > pos
            memo[key] = res, endpos
        return res

    return memoize_wrapper


def memoize_left_rec(func):
    """Memoize a left-recursive parsing method.

    This is similar to @memoize but loops until no longer parse is obtained. Inspired by
    https://github.com/PhilippeSigaud/Pegged/wiki/Left-Recursion
    """

    def memoize_left_rec_wrapper(self, *args):
        pos = self.mark()
        memo = self.memos.get(pos)
        if memo is None:
            memo = self.memos[pos] = {}
        key = (func, args)
        if key in memo:
            res, endpos = memo[key]
            self.reset(endpos)
        else:
            # This is where we deviate from @memoize.

            # Prime the cache with a failure.
            memo[key] = lastres, lastpos = None, pos

            # Loop until no longer parse is obtained.
            while True:
                self.reset(pos)
                res = func(self, *args)
                endpos = self.mark()
                if endpos <= lastpos:
                    break
                memo[key] = lastres, lastpos = res, endpos

            res = lastres
            self.reset(lastpos)

        return res

    return memoize_left_rec_wrapper



"""Test cases for copying models."""

import copy
import numpy as np
import pytest
from ....util import (
    assert_cavities_equivalent,
    assert_gauss_commands_equivalent,
    assert_model_elements_equivalent,
    assert_detectors_equivalent,
)
from finesse.parameter import ParameterState


@pytest.fixture(scope="module")
def model_reference_expression():
    import finesse

    kat = finesse.Model()
    kat.parse(
        """
        l L0 P=1
        s s1 L0.p1 ITM.p1
        m ITM R=1-&ETM.T T=&ETM.T
        s sCAV ITM.p2 ETM.p1 L=1
        m ETM R=0.99 T=0.01

        xaxis(ETM.phi, lin, 0, 1, 100)
        """
    )
    return kat


def test_deepcopy_ref_expr(model_reference_expression):
    model_reference_expression.deepcopy()


@pytest.fixture(scope="module")
def deep_copied_model(complicated_ifo_for_copying):
    """Deep copied complicated model.

    Tests that use this fixture rely on all of the checked parameters NOT BEING EMPTY. That means
    the model here should have components, cavities and detectors.
    """
    if (
        not complicated_ifo_for_copying.components
        or not complicated_ifo_for_copying.cavities
        or not complicated_ifo_for_copying.detectors
    ):
        raise ValueError(
            "deep_copied_model requires a model with components, cavities and detectors"
        )

    return complicated_ifo_for_copying, copy.deepcopy(complicated_ifo_for_copying)


def test_deep_copied_model_ids_dont_match(deep_copied_model):
    """Check that deep copied model IDs are different."""
    original, copied = deep_copied_model
    assert id(original) != id(copied)


def test_deep_copied_model_element_symbols(deep_copied_model):
    A, B = deep_copied_model
    for a, b in zip(A.elements.values(), B.elements.values()):
        for pa, pb in zip(a.parameters, b.parameters):
            if pa.state == ParameterState.Symbolic:
                # if symbolic make sure refs point to new refs
                for sa, sb in zip(pa.value.parameters(), pb.value.parameters()):
                    assert sa.owner is not sb.owner
            else:
                assert pa.value == pb.value


def test_deep_copied_model_have_same_properties(deep_copied_model):
    """Check that deep copied models have the same properties."""
    original, copied = deep_copied_model
    assert np.all(original.homs == copied.homs)
    assert original.phase_config == copied.phase_config
    assert original.lambda0 == copied.lambda0
    assert original.epsilon == copied.epsilon


def test_deep_copied_model_network_objects_dont_match(deep_copied_model):
    """Check that deep copied model networks are not the same object."""
    original, copied = deep_copied_model
    assert original.network != copied.network


def test_deep_copied_model_networks_have_same_properties(deep_copied_model):
    """Check that deep copied model networks have the same properties."""
    original, copied = deep_copied_model
    assert list(original.network) == list(copied.network)
    assert list(original.network.edges) == list(copied.network.edges)


def test_deep_copied_model_optical_network_objects_match(deep_copied_model):
    """Check that deep copied model optical networks are not the same object."""
    original, copied = deep_copied_model
    assert original.optical_network != copied.optical_network


def test_deep_copied_model_optical_networks_have_same_properties(deep_copied_model):
    """Check that deep copied model optical networks have the same properties."""
    original, copied = deep_copied_model
    assert list(original.optical_network) == list(copied.optical_network)
    assert list(original.optical_network.edges) == list(copied.optical_network.edges)


def test_deep_copied_model_cavities_objects_dont_match(deep_copied_model):
    """Check that deep copied model cavities are not the same object."""
    original, copied = deep_copied_model
    assert original.cavities != copied.cavities


def test_deep_copied_model_cavities_have_same_properties(deep_copied_model):
    """Check that deep copied model cavities have the same properties."""
    original, copied = deep_copied_model
    for original_cavity, copied_cavity in zip(original.cavities, copied.cavities):
        assert_cavities_equivalent(original_cavity, copied_cavity)


def test_deep_copied_model_gauss_commands_objects_dont_match(deep_copied_model):
    """Check that deep copied model gauss commands are not the same object."""
    original, copied = deep_copied_model
    assert original.gauss_commands != copied.gauss_commands


def test_deep_copied_model_gauss_commands_have_same_properties(deep_copied_model):
    """Check that deep copied model gauss commands have the same properties."""
    original, copied = deep_copied_model
    for original_gauss, copied_gauss in zip(
        original.gauss_commands, copied.gauss_commands
    ):
        assert_gauss_commands_equivalent(
            original.gauss_commands[original_gauss], copied.gauss_commands[copied_gauss]
        )


def test_deep_copied_model_components_objects_dont_match(deep_copied_model):
    """Check that deep copied model components are not the same object."""
    original, copied = deep_copied_model
    assert original.components != copied.components


def test_deep_copied_model_components_have_same_properties(deep_copied_model):
    """Check that deep copied model components have the same properties."""
    original, copied = deep_copied_model
    for original_component, copied_component in zip(
        original.components, copied.components
    ):
        assert_model_elements_equivalent(original_component, copied_component)


def test_deep_copied_model_detectors_objects_dont_match(deep_copied_model):
    """Check that deep copied model detectors are not the same object."""
    original, copied = deep_copied_model
    assert original.detectors != copied.detectors


def test_deep_copied_model_detectors_have_same_properties(deep_copied_model):
    """Check that deep copied model detectors have the same properties."""
    original, copied = deep_copied_model
    for original_detector, copied_detector in zip(original.detectors, copied.detectors):
        assert_detectors_equivalent(original_detector, copied_detector)


def test_deep_copied_model_elements_objects_dont_match(deep_copied_model):
    """Check that deep copied model elements are not the same objects."""
    original, copied = deep_copied_model
    assert original.elements != copied.elements


def test_deep_copied_model_elements_have_same_properties(deep_copied_model):
    """Check that deep copied model elements have the same properties."""
    original, copied = deep_copied_model
    # Check element names are equal.
    for original_element, copied_element in zip(original.elements, copied.elements):
        # FIXME: check objects within the elements dict too.
        assert original_element == copied_element



"""Test cases for reflected and transmitted power from a single mirror."""

import math
import pytest
from finesse.analysis import noxaxis
from finesse.components import Laser, Space, Mirror
import finesse.detectors as detectors


@pytest.fixture
def laser_and_mirror(model):
    """Model with a laser and mirror separated by a space."""
    model.chain(Laser("L0"), Space("s0"), Mirror("M", R=0.5, T=0.5))
    model.add(detectors.PowerDetector("refl", model.M.p1.o))
    model.add(detectors.PowerDetector("trns", model.M.p2.o))
    return model


def test_fully_reflective(laser_and_mirror):
    """Test fully reflective mirror power."""
    laser_and_mirror.M.set_RTL(R=1, T=0)
    assert float(laser_and_mirror.M.T) == 0

    out = noxaxis(laser_and_mirror)
    assert math.isclose(out["refl"], float(laser_and_mirror.L0.P))
    assert out["trns"] == 0


def test_fully_transmissive(laser_and_mirror):
    """Test fully transmissive mirror power."""
    laser_and_mirror.M.set_RTL(T=1, R=0)
    assert float(laser_and_mirror.M.R) == 0

    out = noxaxis(laser_and_mirror)

    assert math.isclose(out["trns"], float(laser_and_mirror.L0.P))
    assert out["refl"] == 0


def test_half_reflective(laser_and_mirror):
    """Test half reflective mirror power."""
    laser_and_mirror.M.set_RTL(0.5, 0.5)

    assert math.isclose(float(laser_and_mirror.M.T), 0.5)

    out = noxaxis(laser_and_mirror)

    assert math.isclose(out["refl"], 0.5 * float(laser_and_mirror.L0.P))
    assert math.isclose(out["trns"], 0.5 * float(laser_and_mirror.L0.P))



"""
Transmissive optical components which focus or disperse light beams.
"""

import logging

import numpy as np

from finesse.components.modal.lens import LensWorkspace
from finesse.components.general import Connector, InteractionType
from finesse.components.node import NodeDirection, NodeType
from finesse.parameter import float_parameter
from finesse.utilities import refractive_index

LOGGER = logging.getLogger(__name__)


@float_parameter(
    "f",
    "Focal length",
    validate="_check_f",
    units="m",
    is_geometric=True,
)
class Lens(Connector):
    """Represents a thin lens optical component with an associated focal length.

    Notes
    -----
    The specified focal length `f` is only accurate when the lens is attached to spaces with index
    of refraction close to 1. This component exists so that one can use the intuitive focal length
    parameter instead of having to set radii of curvature as with e.g. :class:`.Mirror`.

    Parameters
    ----------
    name : str
        Name of newly created lens.

    f : float, optional
        Focal length of the lens in metres; defaults to infinity.
    """

    def __init__(self, name, f=np.inf):
        super().__init__(name)

        self.f = f

        self._add_port("p1", NodeType.OPTICAL)
        self.p1._add_node("i", NodeDirection.INPUT)
        self.p1._add_node("o", NodeDirection.OUTPUT)

        self._add_port("p2", NodeType.OPTICAL)
        self.p2._add_node("i", NodeDirection.INPUT)
        self.p2._add_node("o", NodeDirection.OUTPUT)

        # optic to optic couplings
        self._register_node_coupling(
            "P1i_P2o",
            self.p1.i,
            self.p2.o,
            interaction_type=InteractionType.TRANSMISSION,
        )
        self._register_node_coupling(
            "P2i_P1o",
            self.p2.i,
            self.p1.o,
            interaction_type=InteractionType.TRANSMISSION,
        )

    def _check_f(self, value):
        if value == 0:
            raise ValueError("Focal length of lens must be non-zero.")

        return value

    def _resymbolise_ABCDs(self):
        self.__symbolise_ABCDs("x")
        self.__symbolise_ABCDs("y")

    def __symbolise_ABCDs(self, direction):
        # TODO (sjr) Not using direction currently but will split
        #            into fx, fy soon to support astigmatic lenses
        M_sym = np.array([[1.0, 0.0], [-1.0 / self.f.ref, 1.0]])

        # Matrices same for both node couplings
        self.register_abcd_matrix(
            M_sym, (self.p1.i, self.p2.o, direction), (self.p2.i, self.p1.o, direction),
        )

    @property
    def abcdx(self):
        """Numeric ABCD matrix in the tangential plane.

        Equivalent to ``lens.ABCD(1, 2, "x")`` and ``lens.ABCD(2, 1, "x")``.

        :getter: Returns a copy of the (numeric) ABCD matrix for this coupling (read-only).
        """
        return self.ABCD(1, 2, "x")

    @property
    def abcdy(self):
        """Numeric ABCD matrix in the sagittal plane.

        Equivalent to ``lens.ABCD(2, 1, "y")`` and ``lens.ABCD(2, 1, "y")``.

        :getter: Returns a copy of the (numeric) ABCD matrix for this coupling (read-only).
        """
        return self.ABCD(1, 2, "y")

    def ABCD(
        self,
        from_node,
        to_node,
        direction="x",
        symbolic=False,
        copy=True,
        retboth=False,
    ):
        r"""Returns the ABCD matrix of the lens for the specified coupling.

        .. _fig_abcd_lens_transmission:
        .. figure:: ../images/abcd_lenst.*
            :align: center

        This is given by,

        .. math::
            M = \begin{pmatrix}
                    1 & 0 \\
                    -\frac{1}{f} & 1
                \end{pmatrix},

        where :math:`f` is the focal length of the lens.

        See :meth:`.Connector.ABCD` for descriptions of parameters, return values and possible
        exceptions.
        """
        return super().ABCD(from_node, to_node, direction, symbolic, copy, retboth)

    def actuate_f(self, dioptres, direction=("x", "y")):
        r"""Actuate on the focal length of the lens with a
        specified dioptre shift.

        Sets the focal length to a new value, :math:`f_2`, via,

        .. math::

            f_2 = \frac{1}{d + \frac{1}{f_1}},

        where :math:`f_1` is the current focal length and :math:`d` is the dioptre shift (i.e.
        the `dioptre` argument).

        By default, both focal planes are shifted. To shift, e.g., only the
        tangential plane, specify ``direction="x"``.

        Parameters
        ----------
        dioptres : float
            Shift in lens focal length in dioptres.

        direction : tuple or str, optional; default: ("x", "y")
            Focal plane to shift, defaults to both tangential and sagittal.
        """
        fnew = lambda x: 1 / (dioptres + 1 / x)
        if "x" in direction:  # TODO (sjr) Change to fx
            self.f = fnew(self.f.value)
        if "y" in direction:  # TODO (sjr) Change to fy
            self.f = fnew(self.f.value)

    def _fill_optical_matrix(self, ws, matrix, connections):
        for freq in matrix.optical_frequencies.frequencies:
            with matrix.component_edge_fill3(
                ws.owner_id, connections.P1i_P2o_idx, freq.index, freq.index,
            ) as mat:
                mat[:] = ws.K12.data

            with matrix.component_edge_fill3(
                ws.owner_id, connections.P2i_P1o_idx, freq.index, freq.index,
            ) as mat:
                mat[:] = ws.K21.data

    def _fill_carrier(self, ws):
        self._fill_optical_matrix(ws, ws.sim.carrier, ws.carrier.connections)

    def _fill_signal(self, ws):
        self._fill_optical_matrix(ws, ws.sim.signal, ws.signal.connections)

    def _get_workspace(self, sim):
        from finesse.simulations.basematrix import CarrierSignalMatrixSimulation

        if isinstance(sim, CarrierSignalMatrixSimulation):
            _, is_changing = self._eval_parameters()
            refill = sim.is_component_in_mismatch_couplings(self) or len(is_changing)

            ws = LensWorkspace(self, sim, refill)
            # This assumes that nr1/nr2 cannot change during a simulation
            ws.nr1 = refractive_index(self.p1)
            ws.nr2 = refractive_index(self.p2)
            # TODO ddb refractive index should be equal on
            # both sides of the lens as we are using the thin
            # lens approximation
            assert ws.nr1 == ws.nr2

            ws.carrier.add_fill_function(self._fill_carrier, refill)
            ws.signal.add_fill_function(self._fill_signal, refill)

            if sim.is_modal:
                ws.abcd_x = self.ABCD(self.p1.i, self.p2.o, "x", copy=False)
                ws.abcd_y = self.ABCD(self.p1.i, self.p2.o, "y", copy=False)

                # Set the coupling matrix information
                # ABCDs are same in each direction
                ws.set_knm_info(
                    "P1i_P2o",
                    abcd_x=ws.abcd_x,
                    abcd_y=ws.abcd_y,
                    nr_from=ws.nr1,
                    nr_to=ws.nr2,
                    is_transmission=True,
                )
                # TODO nr reversed here for now until it's forced to be same on both sides
                ws.set_knm_info(
                    "P2i_P1o",
                    abcd_x=ws.abcd_x,
                    abcd_y=ws.abcd_y,
                    nr_from=ws.nr2,
                    nr_to=ws.nr1,
                    is_transmission=True,
                )

            return ws
        else:
            raise Exception(f"Lens does not handle a simulation of type {sim}")



"""Regular expressions for matching tokens of different types in kat script."""


def group(*choices):
    """Match any of the specified choices."""
    return f"({'|'.join(choices)})"


def any(*choices):
    """Specify that there should be zero or more of the specified choices."""
    return f"{group(*choices)}*"


def maybe(*choices):
    """Specify that there should be zero or one of the specified choices."""
    return f"{group(*choices)}?"


WHITESPACE = r"[ \f\t]+"
NEWLINE = r"\r?\n+"
COMMENT = r"#[^\r\n]*"
NONE = group(r"None", r"none")
BOOLEAN = group(r"True", r"true", r"False", r"false")
REFERENCE = r"&[a-zA-Z_][a-zA-Z0-9_.]*"  # Reference to element name or property.
NAME = r"[a-zA-Z_][a-zA-Z0-9_.]*"  # Element name or property.

INTEGER_NUMBER = r"(?:0(?:_?0)*|[1-9](?:_?[0-9])*)"
EXPONENT = r"[eE][-+]?[0-9](?:_?[0-9])*"
POINT_FLOAT = group(
    r"[0-9](?:_?[0-9])*\.(?:[0-9](?:_?[0-9])*)?", r"\.[0-9](?:_?[0-9])*"
) + maybe(EXPONENT)
EXPONENT_FLOAT = r"[0-9](?:_?[0-9])*" + EXPONENT
FLOAT_NUMBER = group(POINT_FLOAT, EXPONENT_FLOAT)
SI_NUMBER = group(INTEGER_NUMBER, POINT_FLOAT) + r"[pnumkMGT]"
INFINITY = r"inf"
IMAGINARY_NUMBER = group(
    INFINITY + r"[jJ]", r"[0-9](?:_?[0-9])*[jJ]", FLOAT_NUMBER + r"[jJ]"
)
NUMBER = group(IMAGINARY_NUMBER, SI_NUMBER, FLOAT_NUMBER, INTEGER_NUMBER, INFINITY)

SINGLE_QUOTED_STRING = r"'[^\n'\\]*(?:\\.[^\n'\\]*)*'"
DOUBLE_QUOTED_STRING = r'"[^\n"\\]*(?:\\.[^\n"\\]*)*"'
STRING = group(SINGLE_QUOTED_STRING, DOUBLE_QUOTED_STRING)

# Types that only take one form. The order here is important (highest priority first).
LITERALS = {
    "EQUALS": "=",
    "PLUS": "+",
    "MINUS": "-",
    "POWER": "**",  # Has to be above TIMES.
    "TIMES": "*",
    "FLOORDIVIDE": "//",  # Has to be above DIVIDE.
    "DIVIDE": "/",
    "COMMA": ",",
    "LBRACKET": "[",
    "RBRACKET": "]",
    "LPAREN": "(",
    "RPAREN": ")",
}



"""Tests for outputs from asymmetric tracing simulations. Typically checking for
zero scattering into HOMs due to the utility this type of tracing provides."""

import pytest
from finesse.analysis import noxaxis


def test_reflection_from_surface_mismatch(model):
    """Test that mismatch on reflection from a single curved mirror
    is zero when tracing asymmetrically."""
    model.parse(
        """
        l L0 P=1
        s s0 L0.p1 M.p1 L=1
        m M Rc=50

        gauss gL0 L0.p1.o w0=1m z=0

        mmd mmrx M.p1.i M.p1.o x
        mmd mmry M.p1.i M.p1.o y
        """
    )

    model.sim_trace_config["symmetric"] = False
    out = noxaxis(model)

    # Mismatches (in both planes) on reflection from a mirror
    # should be zero when using asymmetric tracing
    assert out["mmrx"] == 0
    assert out["mmry"] == 0


def test_flat_michelson_output_power(model):
    """Test that output power from lossless, fully reflective Michelson with
    flat mirrors is equal to input power (when tuned to light fringe) in the
    case where asymmetric tracing is used. The input beam waist is positioned
    before the Michelson.
    """
    model.parse(
        """
        l L0 P=1
        s s0 L0.p1 BS.p1 L=5

        bs BS

        s sY BS.p2 MY.p1 L=10
        m MY R=1 T=0

        s sX BS.p3 MX.p1 L=10
        m MX R=1 T=0

        gauss gL0 L0.p1.o z=-2.5 w0=1m

        pd P BS.p4.o

        modes(maxtem=0)
        """
    )

    model.sim_trace_config["symmetric"] = False
    out = noxaxis(model)

    # BS tuned to light fringe so fields in phase at output
    # should give P = laser power as mirrors fully reflective
    # -> Only true for asymmetric tracing, as spurious mismatches
    #    are avoided in this case
    assert out["P"] == pytest.approx(model.L0.P.value)
